'''
.. Created on: 2/12/15
   Common functions and data

.. Copyright (c) 2015-2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
import argparse
import os

from .matplotlibFix import plt
from matplotlib.ticker import FuncFormatter

import numpy as np
import seaborn as sns
import shlex

from .error import CommandlineError
from .log import getLogger
from .query import dropExtraCols, readCsv
from .utils import systemOpenFile, pathjoin

_logger = getLogger(__name__)

def setupPalette(count, pal=None):
    # See http://xkcd.com/color/rgb/. These were chosen to be different "enough".
    colors = ['grass green', 'canary yellow', 'dirty pink', 'azure', 'tangerine', 'strawberry',
              'yellowish green', 'gold', 'sea blue', 'lavender', 'orange brown', 'turquoise',
              'royal blue', 'cranberry', 'pea green', 'vermillion', 'sandy yellow', 'greyish brown',
              'magenta', 'silver', 'ivory', 'carolina blue', 'very light brown']

    palette = sns.color_palette(palette=pal, n_colors=count) if pal else sns.xkcd_palette(colors)
    sns.set_palette(palette, n_colors=count)


# For publications, call setupPlot("paper", font_scale=1.5)
def setupPlot(context="talk", style="white", font_scale=1.0):
    sns.set_context(context, font_scale=font_scale)
    sns.set_style(style)


def _getFloatFromFile(filename):
    value = None
    if filename:
        with open(filename) as f:
            value = float(f.readline())

    return value


def _amendFilename(filename, suffix):
    '''
    Insert the given suffix into filename before the extension.
    '''
    base, ext = os.path.splitext(filename)
    return base + '-' + suffix + ext


def _finalizeFigure(fig, ax, outFile=None, yFormat=None, sideLabel=False,
                    labelColor=None, transparent=False, openFile=False, closeFig=True):
    if yFormat:
        func = (lambda x, p: format(int(x), ',')) if yFormat == ',' else (lambda x, p: yFormat % x)
        formatter = FuncFormatter(func)
        ax.get_yaxis().set_major_formatter(formatter)

    if sideLabel:
        labelColor = labelColor or 'lightgrey'
        # add the filename down the right side of the plot
        fig.text(1, 0.5, sideLabel, color=labelColor, weight='ultralight', fontsize=7,
                 va='center', ha='right', rotation=270)

    if outFile:
        fig.savefig(outFile, bbox_inches='tight', transparent=transparent)

    if closeFig:
        plt.close(fig)

    if openFile:
        systemOpenFile(outFile)


def plotUnstackedRegionComparison(df, categoryCol, valueCol=None, region=None,
                                  otherRegion='Rest of world', box=False, title='', ncol=3,
                                  xlabel='', ylabel='', ygrid=False, yticks=False,
                                  ymin=None, ymax=None, legendY=None, palette=None,
                                  outFile=None, sideLabel=False, labelColor=None,
                                  yFormat=None, transparent=False, openFile=False, closeFig=True):
    '''
    Plot unstacked bars showing the values for 'categoryCol', summed across years,
    for one region, for everything else, and the totals of the two.
    '''
    setupPlot()

    count = len(df[categoryCol].unique())     # categoryCol = 'land-allocation'
    setupPalette(count, pal=palette)

    fig, ax = plt.subplots(1, 1, figsize=(8, 4))

    plotCol   = '_value_'
    regionCol = '_region_'

    yearCols = filter(str.isdigit, df.columns)
    if valueCol:
        # Copy value col so we can delete all yearCols
        df[plotCol] = df[valueCol]
    else:
        # Create and plot a new column with values summed across years
        df[plotCol] = df[yearCols].sum(axis=1)

    df = df.drop(yearCols, axis=1)          # copy to not affect caller's df

    regions = region.split(',')
    reg   = df.query('region in %s' % regions)
    other = df.query('region not in %s' % regions)

    grp = other.groupby(categoryCol)
    otherSums = grp.sum()
    otherSums.reset_index(inplace=True)
    otherSums[regionCol] = otherRegion

    grp = reg.groupby(categoryCol)
    regSum = grp.sum()
    regSum.reset_index(inplace=True)
    regSum[regionCol] = region

    # combine the data for selected region and Rest of World, and sum these
    # to create a Total bar
    totals = regSum.set_index(categoryCol) + otherSums.set_index(categoryCol)
    totals[regionCol] = 'Total'
    totals.reset_index(inplace=True)
    world = regSum.append(otherSums).append(totals)

    ax = sns.barplot(x=regionCol, y=plotCol, hue=categoryCol, data=world, ci=None)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    legendY = -0.45 if legendY is None else legendY
    ax.legend(loc='lower center', bbox_to_anchor=(0.5, legendY), ncol=ncol)

    if box == False:
        sns.despine(left=True)

    if yticks:
        plt.tick_params(axis='y', direction='out', length=5, width=.75,
                        colors='k', left='on', right='off')

    if ygrid:
        ax.yaxis.grid(color='lightgrey', linestyle='solid')

    #plt.xticks(rotation=rotation)

    if title:
        ax.set_title(title, y=1.05)

    if ymin is not None or ymax is not None:
        ax.set_autoscale_on(False)
        ax.set_ylim(ymin, ymax)

    _finalizeFigure(fig, ax, outFile=outFile, sideLabel=sideLabel, labelColor=labelColor,
                    yFormat=yFormat, transparent=transparent, openFile=openFile, closeFig=closeFig)

    return (fig, ax)


def plotStackedBarsScalar(df, indexCol, columns, valuesCol, box=False, rotation=90,
                          zeroLine=False, title="", xlabel='', ylabel='', ncol=5, ygrid=False,
                          yticks=False, ymin=None, ymax=None, barWidth=0.5, legendY=None,
                          palette=None, outFile=None, sideLabel=False, labelColor=None,
                          yFormat=None, transparent=False, openFile=False, closeFig=True):
    '''
    Plot a stacked bar plot using data in df, given the index column, the
    column holding the values to pivot to columns, and the column holding
    the values. The argument 'ncol' specifies the number of columns with
    which to render the legend.
    '''
    #_logger.debug('plotStackedBarsScalar %s', sideLabel)
    setupPlot()

    colList = filter(None, [indexCol, columns, valuesCol])  # if indexCol is None, this drops it

    # TBD: handle year values as columns to plot
    df2 = df[colList].pivot(index=indexCol, columns=columns, values=valuesCol)

    setupPalette(len(df2.columns), pal=palette)

    fig, ax = plt.subplots(1, 1, figsize=(8, 4))
    df2.plot(kind='bar', stacked=True, ax=ax, grid=False, width=barWidth, rot=rotation)

    if box == False:
        sns.despine(left=True)

    if yticks:
        plt.tick_params(axis='y', direction='out', length=5, width=.75,
                        colors='k', left='on', right='off')

    if zeroLine:
        ax.axhline(0, color='k', linewidth=0.75, linestyle='-')

    if ygrid:
        ax.yaxis.grid(color='lightgrey', linestyle='solid')

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    legendY = -0.6 if legendY is None else legendY
    ax.legend(loc='upper center', bbox_to_anchor=(0.5, legendY), ncol=ncol)

    if title:
        ax.set_title(title, y=1.05)

    if ymin is not None or ymax is not None:
        ax.set_autoscale_on(False)
        ax.set_ylim(ymin, ymax)

    _finalizeFigure(fig, ax, outFile=outFile, sideLabel=sideLabel, labelColor=labelColor,
                    yFormat=yFormat, transparent=transparent, openFile=openFile, closeFig=closeFig)

    return (fig, ax)


def plotStackedTimeSeries(df, index='region', xlabel='', ylabel='', ncol=5, box=False,
                          zeroLine=False, title="", ygrid=False, yticks=False,
                          ymin=None, ymax=None, barWidth=0.5, legendY=None, yearStep=5,
                          palette=None, outFile=None, sideLabel=False, labelColor=None,
                          yFormat=None, transparent=False, openFile=False, closeFig=True):
    #_logger.debug('plotStackedTimeSeries %s', sideLabel)
    setupPlot()
    df = dropExtraCols(df, inplace=False)
    grouped = df.groupby(index)
    df2 = grouped.aggregate(np.sum)
    df3 = df2.transpose()

    setupPalette(len(df3.columns), pal=palette)
    fig, ax = plt.subplots(1, 1, figsize=(8, 4))
    df3.plot(kind='bar', stacked=True, ax=ax, grid=False, width=barWidth)

    # space out year labels to every 5 years
    locs, labels = plt.xticks()
    yearCols = filter(str.isdigit, df.columns)

    if int(yearCols[1]) - int(yearCols[0]) == 1 and yearStep > 1:
        plt.xticks(locs[::yearStep], yearCols[::yearStep])

    if box == False:
        sns.despine(left=True)

    if yticks:
        plt.tick_params(axis='y', direction='out', length=5, width=.75,
                        colors='k', left='on', right='off')

    lines = ax.get_lines()
    if lines:
        lines[0].set_visible(False)    # get rid of ugly dashed line

    if zeroLine:
        ax.axhline(0, color='k', linewidth=0.75, linestyle='-')

    if ygrid:
        ax.yaxis.grid(color='lightgrey', linestyle='solid')

    if ymin is not None or ymax is not None:
        ax.set_autoscale_on(False)
        ax.set_ylim(ymin, ymax)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    legendY = -0.2 if legendY is None else legendY
    ax.legend(loc='upper center', bbox_to_anchor=(0.5, legendY), ncol=ncol)

    if title:
        ax.set_title(title, y=1.05)

    _finalizeFigure(fig, ax, outFile=outFile, sideLabel=sideLabel, labelColor=labelColor,
                    yFormat=yFormat, transparent=transparent, openFile=openFile, closeFig=closeFig)

    return (fig, ax)


def plotTimeSeries(df, xlabel='', ylabel='', box=False, zeroLine=False, title="", ygrid=False,
                   yticks=False, ymin=None, ymax=None, legend=False, legendY=None, yearStep=5,
                   outFile=None, sideLabel=False, labelColor=None, yFormat=None, transparent=False,
                   openFile=False, closeFig=True):

    setupPlot()
    fig, ax = plt.subplots(1, 1, figsize=(8, 4))

    yearCols = filter(str.isdigit, df.columns)
    x = map(int, yearCols)
    y = list(df[yearCols].iloc[0])
    plt.plot(x, y)

    # TBD: see if this is worth doing
    # space out year labels to every 5 years
    #locs, labels = plt.xticks()
    #plt.xticks(locs[::yearStep], yearCols[::yearStep])

    if box == False:
        sns.despine(left=True)

    if yticks:
        plt.tick_params(axis='y', direction='out', length=5, width=.75,
                        colors='k', left='on', right='off')

    if zeroLine:
        ax.axhline(0, color='k', linewidth=0.75, linestyle='-')

    if ygrid:
        ax.yaxis.grid(color='lightgrey', linestyle='solid')

    if ymin is not None or ymax is not None:
        ax.set_autoscale_on(False)
        ax.set_ylim(ymin, ymax)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    if legend:
        legendY = -0.2 if legendY is None else legendY
        ax.legend(loc='upper center', bbox_to_anchor=(0.5, legendY))
    else:
        ax.legend([], frameon=False)

    if title:
        ax.set_title(title, y=1.05)

    _finalizeFigure(fig, ax, outFile=outFile, sideLabel=sideLabel, labelColor=labelColor,
                    yFormat=yFormat, transparent=transparent, openFile=openFile, closeFig=closeFig)

    return (fig, ax)


def chartGCAM(args, num=None, negate=False):
    """
    Generate a chart from GCAM data. This function is called to process
    the ``chart`` sub-command for a single scenario. See the command-line
    arguments to the ``chart`` sub-command for details about `args`.

    :param args: (argparse Namespace) command-line arguments to `chart`
        sub-command
    :param num: (int or None) if not None, a number to prepend to the
        filename to allow files to have numerical sequence.
    :param negate: (bool) if True, all values in year columns are multiplied
        by -1 before plotting.
    :return: none
    """
    barWidth   = args.barWidth
    box        = args.box
    byRegion   = args.byRegion
    columns    = args.columns
    constraint = args.constraint
    csvFile    = args.csvFile
    indexCol   = args.indexCol or None
    label      = args.label
    labelColor = args.labelColor
    legendY    = args.legendY
    ncol       = args.ncol
    openFile   = args.open
    outFile    = args.outFile
    outputDir  = args.outputDir
    palette    = args.palette
    region     = args.region
    rotation   = args.rotation
    sumYears   = args.sumYears
    timeseries = args.timeseries
    title      = args.title
    transparent= args.transparent
    unstackCol  = args.unstackedCol
    unstackReg = args.unstackedRegion
    valueCol   = args.valueCol
    xlabel     = args.xlabel
    yFormat    = args.format
    yearStep   = args.yearStep
    ygrid      = args.ygrid
    ylabel     = args.ylabel
    ymax       = args.ymax
    ymin       = args.ymin
    yticks     = args.yticks
    zeroLine   = args.zeroLine

    # DOCUMENT
    # use outputDir if provided, else use parent dir of outFile
    outputDir = outputDir or os.path.dirname(outFile)

    if not os.path.lexists(outputDir):
        os.mkdir(outputDir, 0o755)

    if outFile:
        imgFile = os.path.basename(outFile)
    else:
        # otherwise compute output filename using outputDir and suffix
        suffix = args.suffix or '.png'
        suffix = '-' + suffix if suffix[0] not in ['.', '-', '_'] else suffix
        filename = os.path.basename(csvFile)
        root, ext = os.path.splitext(filename)
        prefix = "%d-" % num if num else ""
        imgFile = prefix + root + suffix

    outFile = pathjoin(outputDir, imgFile)

    _logger.debug("Generating %s", os.path.abspath(outFile))

    if args.years:
        yearStrs = args.years.split('-')
        assert len(yearStrs) == 2, "Year range must be given as YYYY-YYYY"
    else:
        yearStrs = None

    # e.g., "/Users/rjp/ws-ext/new-reference/batch-new-reference/LUC_Emission_by_Aggregated_LUT_EM-new-reference." % scenario
    df = readCsv(csvFile, skiprows=args.skiprows, years=yearStrs, interpolate=args.interpolate)

    if region:
        try:
            df = df.query('region == "%s"' % region)
        except Exception as e:
            raise CommandlineError("Failed to slice by region %s\n  -- %s" % (region, e))

        if df.shape[0] == 0:
            raise CommandlineError('Region "%s" was not found in %s' % (region, csvFile))

    if constraint:
        try:
            df = df.query(constraint)
        except Exception as e:
            raise CommandlineError("Failed to apply constraint: %s\n  -- %s" % (constraint, e))

    yearCols = filter(str.isdigit, df.columns)

    if args.multiplier or args.divisor or negate:
        df = df.copy(deep=True)

    multiplier = args.multiplier or _getFloatFromFile(args.multiplierFile)
    if multiplier:
        df[yearCols] *= multiplier

    divisor = args.divisor or _getFloatFromFile(args.divisorFile)
    if divisor:
        df[yearCols] /= divisor

    if negate:
        outFile = _amendFilename(outFile, 'negated')
        imgFile = _amendFilename(imgFile, 'negated')
        df[yearCols] *= -1

    regions = df.region.unique() if byRegion else [None]    # allows loop to work when not byRegion

    outFileOrig = outFile
    imgFileOrig = imgFile
    titleOrig   = title
    dfOrig = df

    for reg in regions:
        if reg:
            df = dfOrig.query('region == "%s"' % reg)
            title = titleOrig + " (%s)" % reg
            outFile = _amendFilename(outFileOrig, reg)
            imgFile = _amendFilename(imgFileOrig, reg)
            _logger.debug("Processing %s", reg)

        sideLabel = imgFile if label else ''

        if unstackCol:
            otherRegion = 'Rest of world'
            mainRegion  = reg or unstackReg

            plotUnstackedRegionComparison(df, unstackCol, valueCol=valueCol, region=mainRegion,
                                          otherRegion=otherRegion, box=box, title=title, ncol=ncol,
                                          xlabel=xlabel, ylabel=ylabel, ygrid=ygrid, yticks=yticks,
                                          ymin=ymin, ymax=ymax, legendY=legendY, palette=palette,
                                          outFile=outFile, sideLabel=sideLabel, labelColor=labelColor,
                                          yFormat=yFormat, transparent=transparent, openFile=openFile)
        elif sumYears or valueCol:
            if sumYears:
                # create a new value column by summing year columns
                valueCol = '_total_'
                df[valueCol] = df[yearCols].sum(axis=1)

            plotStackedBarsScalar(df, indexCol, columns, valueCol, box=box, zeroLine=zeroLine,
                                  title=title, xlabel=xlabel, ylabel=ylabel, ygrid=ygrid, yticks=yticks,
                                  ymin=ymin, ymax=ymax, rotation=rotation, ncol=ncol, barWidth=barWidth,
                                  legendY=legendY, palette=palette, outFile=outFile, sideLabel=sideLabel,
                                  labelColor=labelColor, yFormat=yFormat, transparent=transparent,
                                  openFile=openFile)

        elif timeseries:
            plotTimeSeries(df, xlabel=xlabel, ylabel=ylabel, box=box, zeroLine=zeroLine, title=title, ygrid=ygrid,
                           yticks=yticks, ymin=ymin, ymax=ymax, legend=False, legendY=legendY, yearStep=yearStep,
                           outFile=outFile, sideLabel=sideLabel, labelColor=labelColor, yFormat=yFormat,
                           transparent=transparent, openFile=openFile)

        else:
            plotStackedTimeSeries(df, index=indexCol, yearStep=yearStep, ygrid=ygrid, yticks=yticks,
                                  ymin=ymin, ymax=ymax, zeroLine=zeroLine, title=title, legendY=legendY,
                                  box=box, xlabel=xlabel, ylabel=ylabel, ncol=ncol, barWidth=barWidth,
                                  palette=palette, outFile=outFile, sideLabel=sideLabel, labelColor=labelColor,
                                  yFormat=yFormat, transparent=transparent, openFile=openFile)

def chartMain(mainArgs, tool, parser):
    # DOCUMENT '*null*', if still useful
    if not mainArgs.fromFile and mainArgs.csvFile == '*null*':
        raise CommandlineError("Must specify a CSV file or use -f flag to read arguments from a file")

    os.chdir(mainArgs.workingDir)

    # Process these separately
    negate = mainArgs.negate
    del mainArgs.negate

    if mainArgs.fromFile:
        del mainArgs.csvFile    # whatever is passed, e.g., "-", is ignored

        enumerate = mainArgs.enumerate
        num = 1

        with open(mainArgs.fromFile) as f:
            lines = f.readlines()

        # Set up dictionary for substitution into lines as they're read
        substDict = {'scenario'  : None,    # set to current scenario below
                     'reference' : mainArgs.reference,
                     'years'     : mainArgs.years}

        scenarios = mainArgs.scenario.split(',')

        for scenario in scenarios:
            substDict['scenario'] = scenario
            argDict = vars(mainArgs)
            argDict['scenario'] = scenario  # for each call, pass the current scenario only

            # Merge command-line args with those from the file,
            # allowing the ones from the file to override.
            for line in lines:
                # ignore comment lines
                line = line.strip()
                if not line or line[0] == '#':
                    continue

                if line == 'exit':
                    return

                line = line.format(**substDict)
                fileArgs = shlex.split(line)

                argsNS  = argparse.Namespace(**argDict)
                allArgs = parser.parse_args(args=fileArgs, namespace=argsNS)

                nextNum = num if enumerate else None
                num += 1

                chartGCAM(allArgs, num=nextNum, negate=negate)

    else:
        chartGCAM(mainArgs, negate=negate)




from __future__ import print_function
from collections import OrderedDict

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

from pygcam.tool import GcamTool
from pygcam.log import getLogger
from pygcam.subcommand import SubcommandABC
from pygcam.gui.actions import actionTable, getActionInstance

_logger = getLogger(__name__)

def dataStore(id):
    """
    Generate an invisible div with the given id
    """
    return html.Div(id=id, style={'display': 'none'})

def subcommandHelp(cmd):
    tool = GcamTool.getInstance(reload=True)
    cmds = tool.subparsers._choices_actions
    help = next((s.help for s in cmds if s.dest == cmd), '')
    return help

def actionsFromParser(cmd, parser):
    actions = map(lambda action: getActionInstance(cmd, action), parser._actions)
    return filter(None, actions)    # remove Nones

#
# Class to support multi-page applications with dash
#
class Page(object):
    """
    Defines one app page, which may support a set of subpages
    """
    def __init__(self, app, id, label=None, heading=None, pageSet=None):
        self.app = app
        self.id = id
        self.pageSet = pageSet
        self.actions = None
        self.helpText = None
        self.heading = heading or 'Options for "%s"' % id
        self.label = label or id.capitalize()

        if pageSet:
            return

        subcmd = SubcommandABC.getInstance(id)
        if subcmd:
            parser = subcmd.parser
            self.actions = actionsFromParser(id, parser)
            self.helpText = subcommandHelp(id)
            if not label:
                self.label = subcmd.label

        self.generateCallbacks() # callbacks for generated widgets

    def __str__(self):
        return "<Page id='%s' label='%s'>" % (self.id, self.label)

    def render(self):
        pageSet = self.pageSet
        layout = self.getLayout()
        layout = html.Div([pageSet.render(), layout]) if pageSet else layout
        return layout

    def select(self, id):
        return self.pageSet.select(id) if self.pageSet else None

    def getLayout(self):
        if not self.actions:
            return ''

        # Generate rather than saving this since we want to maintain state across virtual pages
        return html.Div(actionTable(self.heading, self.actions, helpText=self.helpText))

    def getArgs(self):
        args = [action.cmdlineArg() for action in self.actions]
        args = filter(None, args)   # remove Nones
        return ' '.join(args)

    def getCommand(self):
        """
        Return the command implied by the values in the GUI
        """
        globalArgs = RootPage.globalArgs()
        localArgs = self.getArgs()
        args = ('gt', globalArgs, self.id, localArgs)
        cmd  = ' '.join(args)
        return cmd

    def generateCallbacks(self):
        app = self.app
        if self.actions:
            for action in self.actions:
                action.generateCallback(app)


class GlobalArgsPage(Page):
    def __init__(self, app, id='globalArgs'):
        super(GlobalArgsPage, self).__init__(app, id=id, label='Global args', heading='Global arguments')
        self.helpText = 'The options on this page apply to all gt commands'
        tool = GcamTool.getInstance()
        self.actions = actionsFromParser(id, tool.parser)
        self.generateCallbacks() # callbacks for generated widgets

class PageSet(object):
    def __init__(self, id, pages, default):
        self.id = id
        self.contentId = id + '-content'

        self.pages = OrderedDict()
        for page in pages:
            self.pages[page.id] = page

        self.default = default
        self.selected = self.select(self.default)
        self.navPrefix = 'sub-nav-'  # non-Root (lower-level) menu items


    def __str__(self):
        return "<PageSet %s default:%s selected:%s>" % (self.pages.keys(), self.default, self.selected)

    def select(self, id):
        self.selected = self.pages[id or self.default]
        return self.selected

    def navbar(self):
        pages = self.pages.values()
        prefix = self.navPrefix

        def pageURL(page):
            return "/%s/%s" % (self.id, page.id) if self.id else '/' + page.id

        def buttonClass(page):
            return prefix + ('selected' if page == self.selected else 'button')

        layout = html.Div([
            html.Div([dcc.Link(pg.label, href=pageURL(pg), className=buttonClass(pg)) for pg in pages],
                     className=(prefix + 'bar'))],
            className=(prefix + 'bg'))
        return layout

    def render(self):
        page = self.selected
        contents = page.render() if page else ''
        layout = html.Div([self.navbar(), contents])
        return layout


class RootPage(PageSet):
    instance = None

    def __init__(self, app, term, pages, default=None):
        RootPage.instance = self

        url = dcc.Location(id='url', refresh=False)
        app.layout = html.Div([url,
                               html.Div(id='root-content'),
                               html.Div([html.H3('Command terminal'),
                                         term.layout],
                                        id='terminal-div')
                               ])

        self.navPrefix = 'nav-'  # Root (top-level) menu items (must set after super.__init__)
        default = default or pages[0].id

        super(RootPage, self).__init__('', pages, default)

        @app.callback(Output('root-content', 'children'),
                      [Input('url', 'pathname')])
        def displayPage(pathname):
            _logger.debug("Pathname is", pathname)
            id = pathname[1:] if pathname else self.default
            elts = id.split('/')
            pageId = elts[0]
            page = self.select(pageId)

            if page and len(elts) > 1:
                subpage = elts[1]
                page.select(subpage)

            layout = self.render()
            selected = page.pageSet.selected if page.pageSet else page

            term.setPage(selected)
            return [layout]

        @app.callback(Output('terminal-div', 'style'),
                      [Input('url', 'pathname')])
        def showTerminal(pathname):
            # we don't show terminal for globalArgs since it's not a sub-command
            value = 'none' if pathname == '/globalArgs' else 'inline'
            style = {'display': value}
            return style

    @classmethod
    def globalArgs(cls):
        self = RootPage.instance
        globalArgsPage = self.pages['globalArgs']
        return globalArgsPage.getArgs()



#!/usr/bin/env python
#
import sys
from os.path import dirname, realpath

XmlSourceDir = dirname(realpath(sys.argv[0]))
sys.path.insert(0, dirname(dirname(dirname(XmlSourceDir))))

from pygcam.scenarioSetup import ConfigEditor
from pygcam.sectorEditors import RefiningEditor, BioenergyEditor

# Paper1.scenarios import (BiofuelShock, ResultsDir, XmlOutputRoot,
#                                  cornName, cornTag, cellName, cellTag,
#                                  bdName, bdTag, ftName, ftTag)
# from xmlsrc.lib.BaseScenario import parseArgs

# from xmlsrc.common.genDeltaConstraints import genDeltaConstraints

# This sets default baseline and scenario, both of which can be
# overridden on the command-line to scenarioSetup.py
args = ConfigEditor.parseArgs(baseline='base-1', scenario='corn-1')

BaselineName  = args.baseline
ScenarioName  = args.scenario
Years         = args.years
resultsDir    = args.resultsDir # or ResultsDir
generate      = not args.noGenerate
#xmlOutputRoot = args.xmlOutputRoot or XmlOutputRoot

cornPolicyType = 'subsidy'
cellPolicyType = 'subsidy'
ftPolicyType   = 'subsidy'
bdPolicyType   = 'subsidy'

cornTag = 'corn ethanol'
cellTag = 'cellulosic ethanol'
ftTag   = 'FT biofuels'
bdTag   = 'biodiesel'

UseDynXml = False

class BiofuelShock(ConfigEditor, RefiningEditor, BioenergyEditor):
    def __init__(self, *args, **kwargs):
        super(BiofuelShock, self).__init__(*args, **kwargs)

s = BiofuelShock(ScenarioName, BaselineName, XmlOutputRoot, XmlSourceDir)
s.setup(dynamic=UseDynXml)
s.setSolutionTolerance(0.01)

s.addMarketConstraint(cornTag, cornPolicyType, dynamic=UseDynXml)
s.addMarketConstraint(cellTag, cellPolicyType, dynamic=UseDynXml)
s.addMarketConstraint(ftTag,   ftPolicyType,   dynamic=UseDynXml)
s.addMarketConstraint(bdTag,   bdPolicyType,   dynamic=UseDynXml)



# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from pygcam.log import getLogger
from .McsSubcommandABC import McsSubcommandABC

_logger = getLogger(__name__)

def driver(args, tool):
    """
    Analyze MCS results
    """
    import os
    from ..analysis import analyzeSimulation
    from ..error import PygcamMcsUserError

    if args.timeseries:
        import pandas as pd
        from pygcam.config import getParam
        from ..Database import getDatabase
        from ..timeseriesPlot import plotTimeSeries, plotForcingSubplots
        from ..util import stripYearPrefix

        simId = args.simId
        expList = args.expName.split(',')
        resultName = args.resultName
        xlabel = 'Year' # args.xlabel or 'Year'
        ymin = args.ymin
        ymax = args.ymax

        # special purpose plot
        forcingPlot = args.forcingPlot

        if not (expList[0] and resultName):
            raise PygcamMcsUserError("expName and resultName must be specified")

        db = getDatabase()
        trialCount = db.getTrialCount(simId)

        plotDir  = getParam('MCS.PlotDir')
        plotType = getParam('MCS.PlotType')

        allResults = db.getTimeSeries(simId, resultName, expList) # , regionName)
        if not allResults:
            raise PygcamMcsUserError('No timeseries results for simId=%d, expList=%s, resultName=%s' \
                                     % (simId, expList, resultName))

        def computeFilename(expName):
            basename = "%s-s%d-%s.%s" % (resultName, simId, expName, plotType)
            filename = os.path.join(plotDir, 's%d' % simId, basename)
            return filename

        # massage the data into the format required by plotTimeSeries
        def createRecord(pair):
            obj, expName = pair
            d = obj.__dict__
            d['expName'] = expName
            return d

        records = [createRecord(pair) for pair in allResults]
        resultDF = pd.DataFrame.from_records(records, index='seriesId')
        units = resultDF.units.iloc[0]
        resultDF.drop(['units', '_sa_instance_state', 'outputId'], axis=1, inplace=True)

        # convert column names like 'y2020' to '2020'
        cols = map(stripYearPrefix, resultDF.columns)
        resultDF.columns = cols

        if forcingPlot:
            filename = computeFilename('combo')
            plotForcingSubplots(resultDF, filename=filename, ci=95, show_figure=False)
            return

        for expName in expList:
            # create a copy so we can drop expName column for melt
            df = resultDF.query("expName == '%s'" % expName).copy()
            _logger.debug("Found %d result records for exp %s, result %s" % (len(df), expName, resultName))

            df.drop(['expName'], axis=1, inplace=True)
            df = pd.melt(df, id_vars=['runId'], var_name='year')

            title = '%s for %s' % (resultName, expName)
            filename = computeFilename(expName)
            _logger.debug('Saving timeseries plot to %s' % filename)

            reg = "" # '-' + regionName if regionName else ""
            extra = "name=%s trials=%d/%d simId=%d scenario=%s%s" % \
                    (resultName, resultDF.shape[0], trialCount, simId, expName, reg)

            # TBD: generalize this with a lookup table or file
            if units == 'W/m^2':
                units = 'W m$^{-2}$'

            plotTimeSeries(df, 'year', 'runId', title=title, xlabel=xlabel, ylabel=units, ci=[95], # [50, 95],
                           text_label=None, legend_name=None, legend_labels=None,
                           ymin=ymin, ymax=ymax, filename=filename, show_figure=False, extra=extra)

        # if doing timeseries plot, none of the other options are relevant
        return

    if not (args.exportInputs or args.resultFile or args.plot or args.importance or
                args.groups or args.plotInputs or args.stats or args.convergence or
                args.exportEMA):
        msg = 'Must specify at least one of: --export, --resultFile, --plot, --importance, --groups, --distros, --stats, --convergence, --exportEMA'
        raise PygcamMcsUserError(msg)

    analyzeSimulation(args)


class AnalyzeCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Analyze simulation results stored in the database for the given simulation.
            At least one of -c, -d, -i, -g, -p, -t (or the longname equivalent) must be specified.'''}
        super(AnalyzeCommand, self).__init__('analyze', subparsers, kwargs)

    def addArgs(self, parser):
        parser.add_argument('-c', '--convergence', action='store_true', default=False,
                            help='Generate convergence plots for mean, std dev, skewness, and 95%% coverage interval.')

        parser.add_argument('-d', '--distros', dest='plotInputs', action='store_true', default=False,
                            help='Plot frequency distributions for input parameters.')

        parser.add_argument('-e', '--expName', type=str,
                            help='The name of the experiment or scenario to run.')

        parser.add_argument('--exportEMA', type=str, default=None,
                            help='''Export results to the given .tar.gz file in a format suitable for analysis
                            using the EMA Workbench. The -e (--expName) and -r (--resultName) flags can hold
                            comma-delimited lists of experiments and results, respectively.''')

        parser.add_argument('--forcingPlot', action='store_true',
                            help='''Plot the data in a good format for multiple forcing timeseries plots''')

        parser.add_argument('-g', '--groups', action='store_true',
                            help='Show the uncertainty importance for groups of parameters.')

        parser.add_argument('-i', '--importance', action='store_true', default=False,
                            help='Show the uncertainty importance for each parameter.')

        parser.add_argument('-l', '--limit', type=int, required=False, default=-1,
                            help='Limit the analysis to the given number of results')

        parser.add_argument('-m', '--min', type=float, required=False, default=None,
                            help='''Limit the analysis to values (for the result named with -r) greater
                            than or equal to this value''')

        parser.add_argument('-M', '--max', type=float, required=False, default=None,
                            help='''Limit the analysis to values (for the result named with -r) less
                            than or equal to this value''')

        parser.add_argument('-o', '--exportInputs', type=str, default=None,
                            help='A file into which to export input (trial) data.')

        parser.add_argument('-O', '--resultFile', type=str, default=None,
                            help='''Export all model results to the given file. When used with this option,
                            the -r (--resultName) and -e (--expName) flags can be comma-delimited lists of
                            result names and experiment names (scenarios), respectively. The output file,
                            in CSV format will have a header (and data in the form) "trialNum,value,expName,resultName"''')

        parser.add_argument('-p', '--plot', action='store_true', default=False,
                            help='''Plot a histogram of the frequency distribution for the named model output
                            (-r required).''')

        parser.add_argument('-R', '--regionName', default=None,
                            help='The region to plot timeseries results for')

        parser.add_argument('-r', '--resultName', type=str, default=None,
                            help='The name of the result variable to analyze.')

        parser.add_argument('-s', '--simId', type=int, default=1,
                            help='The id of the simulation')

        parser.add_argument('-S', '--stats', action='store_true', default=False,
                            help='Print mean, median, max, min, std dev, skewness, and 95%% coverage interval.')

        parser.add_argument('-t', '--timeseries', action='store_true',
                            help='Plot a timeseries distribution')

        parser.add_argument('-x', '--xlabel', dest='xlabel', type=str, default=r'g CO$_2$e MJ$^{-1}$',
                            help='Specify a label for the x-axis in the histogram.')

        parser.add_argument('--ymax', type=float, default=None,
                            help='''Set the scale of a figure by indicating the value to show as the
                            maximum Y value. (By default, scale is set according to the data.)''')

        parser.add_argument('--ymin', type=float, default=None,
                            help='''Set the scale of a figure by indicating the value (given as abs(value),
                            but used as -value) to show as the minimum Y value''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



from __future__ import print_function
from lxml import etree as ET
import os
import pkg_resources as pkg

from pygcam.config import getConfigDict, getParam, stringTrue
from pygcam.log import getLogger
from pygcam.error import XmlFormatError, PygcamException

_logger = getLogger(__name__)

CONDITIONAL = 'CONDITIONAL'
TEST = 'TEST'
THEN = 'THEN'
ELSE = 'ELSE'
OR   = 'OR'
AND  = 'AND'

_ops = {
    '=' : lambda a, b: a == b,
    '==': lambda a, b: a == b,
    '!=': lambda a, b: a != b,
    '<' : lambda a, b: a <  b,
    '<=': lambda a, b: a <= b,
    '>' : lambda a, b: a >  b,
    '>=': lambda a, b: a >= b,

    'eq': lambda a, b: a == b,
    'ne': lambda a, b: a != b,
    'lt': lambda a, b: a <  b,
    'le': lambda a, b: a <= b,
    'gt': lambda a, b: a >  b,
    'ge': lambda a, b: a >= b,
}

_types = {'str': str, 'int': int, 'float': float, 'bool': bool}

# TBD: Modified from version from mcs.XML

class XMLFile(object):
    """
    Stores information about an XML file; provides wrapper to parse and access
    the file tree, and handle "conditional XML".

    :param filename: (str) The pathname to the XML file
    :param load: (bool) If True, the file is loaded, otherwise, the instance is
       set up, but the file is not read.
    :param schemaPath: (str) If not None, the path relative to the root of the
       package to the .xsd (schema definition) file to use to validate the XML file.
    :param removeComments: (bool) If True, comments are discarded upon reading the file.
    :param conditionalXML: (bool) If True, the XML is processed using Conditional XML
       prior to validation.
    :param varDict: (dict) A dictionary to use in place of the configuration dictionary
       when processing Conditional XML.
    """
    def __init__(self, filename, load=True, schemaPath=None,
                 removeComments=True, conditionalXML=False, varDict=None):
        self.filename = filename
        self.tree = None
        self.conditionalXML = conditionalXML
        self.varDict = varDict or getConfigDict(section=getParam('GCAM.DefaultProject'))
        self.removeComments = removeComments

        self.schemaPath   = schemaPath
        self.schemaStream = None

        if filename and load:
            self.read()

    def getRoot(self):
        'Return the root node of the parse tree'
        return self.tree.getroot()

    def getTree(self):
        'Return XML parse tree.'
        return self.tree

    def getFilename(self):
        'Return the filename for this ``XMLFile``'
        return self.filename

    def read(self):
        """
        Read the XML file, and if validate if ``self.schemaFile`` is not None.
        """
        filename = self.filename

        _logger.debug("Reading '%s'", filename)
        parser = ET.XMLParser(remove_blank_text=True, remove_comments=self.removeComments)

        try:
            tree = self.tree = ET.parse(filename, parser)

        except Exception as e:
            raise XmlFormatError("Can't read XML file '%s': %s" % (filename, e))

        if self.conditionalXML:
            self.evaluateConditionals(tree.getroot())

        if self.removeComments:
            for elt in tree.iterfind('//comment'):
                parent = elt.getparent()
                if parent is not None:
                    parent.remove(elt)

        self.validate()

        return tree

    def validate(self, raiseOnError=True):
        """
        Validate a ParameterList against ``self.schemaFile``. Optionally raises an
        error on failure, else return boolean validity status. If no schema file
        is defined, return ``True``.
        """
        if not self.schemaPath:
            return True

        tree = self.tree

        # ensure that the entire directory has been extracted so that 'xs:include' works
        pkg.resource_filename('pygcam', os.path.dirname(self.schemaPath))
        abspath = pkg.resource_filename('pygcam', self.schemaPath)

        xsd = ET.parse(abspath)
        schema = ET.XMLSchema(xsd)

        if raiseOnError:
            try:
                schema.assertValid(tree)
                return True
            except ET.DocumentInvalid as e:
                raise XmlFormatError("Validation of '%s'\n  using schema '%s' failed:\n  %s" % (self.filename, self.schemaPath, e))
        else:
            valid = schema.validate(tree)
            return valid

    def evalTest(self, node):
        tag = node.tag

        if tag == TEST:
            varName  = node.get('var')
            op       = node.get('op', '==')      # defaults to equality
            value    = node.get('value')
            typeName = node.get('type', 'str')   # defaults to str comparison
            varValue = self.varDict.get(varName)
            typeFunc = stringTrue if typeName == 'bool' else _types[typeName]

            def coerce(value, func):
                try:
                    return func(value)
                except Exception:
                    raise XmlFormatError('Failed to convert variable "%s" value "%s" to %s', varName, value, func)

            value    = coerce(value, typeFunc)
            varValue = coerce(varValue, typeFunc)

            try:
                func = _ops[op]
            except KeyError:
                # Shouldn't happen if schema is correct
                raise XmlFormatError('Unknown comparison operator (%s) in conditional XML' % op)

            result = func(varValue, value)
            _logger.debug('<test $%s %s %r> -> %s' % (varName, op, value, result))
            return result

        if tag == AND:
            for child in node:
                if not self.evalTest(child):
                    _logger.debug('<AND> -> False')
                    return False
            _logger.debug('<AND> -> True')
            return True

        if tag == OR:
            for child in node:
                if self.evalTest(child):
                    _logger.debug('<OR> -> True')
                    return True
            _logger.debug('<OR> -> False')
            return False

        raise XmlFormatError('Expected one of %s; got %s' % ((TEST, AND, OR), tag))

    def chooseBranch(self, ifnode):
        tests = ifnode.xpath('%s|%s|%s' % (TEST, AND, OR))

        if len(tests) != 1:
            # Shouldn't happen if schema is correct
            raise XmlFormatError('Expected 1 test|and|or node, got %d' % len(tests))

        test = tests[0]

        result = self.evalTest(test)
        branch = ifnode.find(THEN if result else ELSE)
        return branch

    def evaluateConditionals(self, parent):
        for child in parent:
            if child.tag == CONDITIONAL:
                branch = self.chooseBranch(child)
                if branch is not None:          # test because <else> is optional
                    self.evaluateConditionals(branch)

                    last = child                # insert after the <conditional>
                    for elt in branch:
                        last.addnext(elt)
                        last = elt
                parent.remove(child)            # remove the <conditional>
            else:
                self.evaluateConditionals(child)


class McsValues(XMLFile):
    def __init__(self, filename):
        """
        Reads an XML file of numeric values for named parameters by region.
        This allows the Monte Carlo simulation machinery to modify these
        values. Note that the file's contents are ignored if read by GCAM.

        :param filename: (str) the name of an XML file adhering to mcsValues-schema.xsd.
        """
        from collections import defaultdict

        _logger.debug('Reading MCS values from %s', filename)
        super(McsValues, self).__init__(filename, schemaPath='etc/mcsValues-schema.xsd')

        self.regionMap = defaultdict(dict)
        for regNode in self.tree.iterfind('//region'):
            region = regNode.get('name')
            valueMap = self.regionMap[region]
            for valNode in regNode.iterfind('.//value'):
                name = valNode.get('name')
                valueMap[name] = float(valNode.text)    # xml schema requires numeric value here

    def regions(self):
        """
        Get the regions used in this file.

        :return: (list of str) region names
        """
        return self.regionMap.keys()

    def values(self, region):
        """
        Get the values associated with the `region`.

        :param region: (str) the name of a GCAM region
        :return: (dict) keys are names of value elements and values are floats.
           Returns None if the region is not found
        """
        return self.regionMap.get(region, None)

    def valueForRegion(self, paramName, region, default=None, raiseError=True):
        regionMap = self.values(region)
        if regionMap is None:
            if raiseError:
                raise PygcamException('Region "%s" is not defined in %s' % (region, self.filename))
            return None

        return regionMap.get(paramName, default)

if __name__ == '__main__':
    pathname = '../tests/data/xml/Conditional.xml'
    pathname = '/Users/rjp/bitbucket/paper1/etc/project.xml'
    varDict  = {'bar': 'bar', 'foo': 'foo', 'baz': 20, 'mcsMode': True}
    _logger.setLevel('INFO')

    print('varDict=%s' % varDict)

    xml = XMLFile(pathname, varDict=varDict, schemaPath='etc/project-schema.xsd')
    root = xml.getRoot()

    xml.evaluateConditionals(root)

    print("Result:")
    ET.dump(root)



import os
from tempfile import mkstemp, mkdtemp

from .error import PygcamException
from .config import getParam

def getTempFile(suffix='', tmpDir=None, text=True, delete=True):
    """
    Convenience function for common use pattern, which is to get
    the name of a temp file that needs to be deleted on app exit.

    :param suffix: (str) an extension to give the temporary file
    :param tmpDir: (str) the directory in which to create the file.
      (Defaults to the value of configuration file variable 'GCAM.TempDir',
      or '/tmp' if the variable is not found.
    :param text: True if this will be a text file
    :param delete: (bool) if False, don't delete the file on cleanup.
       (This is useful for debugging.)
    :return: (str) pathname of a new temporary file
    """
    obj = TempFile(suffix=suffix, text=text, tmpDir=tmpDir, delete=delete)
    return obj.path

def getTempDir(suffix='', tmpDir=None, delete=True):
    """
    Convenience function for common use pattern, which is to get the
    name of a temporary directory that needs to be deleted on app exit.

    :param suffix: (str) an extension to give the temporary file
    :param tmpDir: (str) the directory in which to create the new temporary
        directory (Defaults to the value of configuration file variable
        'GCAM.TempDir', or '/tmp' if the variable is not found.
    :param delete: (bool) if False, don't delete the file on cleanup.
       (This is useful for debugging.)
    :return: (str) pathname of a new temporary directory
    """
    obj = TempFile(suffix=suffix, tmpDir=tmpDir, delete=delete, createDir=True)
    return obj.path


class TempFile(object):
    """
    Class to create and track temporary files in one place
    so they can be deleted before an application exits.
    """
    Instances = {}

    def __init__(self, path=None, suffix='', tmpDir=None, delete=True,
                 openFile=False, text=True, createDir=False):
        """
        Construct the name of a temporary file.

        :param path: (str) a path to register for deletion. If given, all other
            args are ignored.
        :param suffix: (str) an extension to give the temporary file
        :param tmpDir: (str) the directory in which to create the (defaults to
            the value of configuration file variable 'GCAM.TempDir', or '/tmp'
            if the variable is not found.
        :param delete: (bool) whether deleteFile() should delete the file when
            called
        :param openFile: (bool) whether to leave the new file open (ignored if
            createDir is True)
        :param text: (bool) Set to False if this will not be a text file
        :param createDir: (bool) if True, a temporary directory will be created
            rather than a temporary file.
        :return: none
        """
        from .utils import mkdirs

        self.suffix = suffix
        self.delete = delete
        self.fd = None

        if path:
            # If we're called with a path, it's just to register a file for deletion.
            # We ignore all other parameters.
            self.path = path
        else:
            tmpDir = tmpDir or getParam('GCAM.TempDir') or "/tmp"
            mkdirs(tmpDir)

            if createDir:
                self.path = mkdtemp(suffix=suffix, dir=tmpDir)
            else:
                fd, tmpFile = mkstemp(suffix=suffix, dir=tmpDir, text=text)

                self.path = tmpFile
                if openFile:
                    self.fd = fd
                else:
                    # the caller is just after a pathname, so close it here
                    os.close(fd)
                    os.unlink(tmpFile)

        # save this instance by the unique path
        self.Instances[self.path] = self

    def deleteFile(self):
        """
        Remove the file for a TempFile instance if ``self.delete`` is True. In either
        case, delete the instance from the class instance dict.

        :return: none
        :raises: PygcamException if the path is not related to a TempFile instance.
        """
        from .log import getLogger
        _logger = getLogger(__name__)

        path = self.path

        try:
            if self.fd is not None:
                os.close(self.fd)
        except Exception as e:
            _logger.debug('Failed to close file descriptor for "%s": %s', path, e)

        try:
            del self.Instances[path]
        except KeyError:
            raise PygcamException('No TempFile instance with name "%s"' % path)

        deleting = 'Deleting' if self.delete else 'Not deleting'
        _logger.debug("%s TempFile file '%s'", deleting, path)

        if self.delete:
            from .utils import removeFileOrTree
            try:
                removeFileOrTree(path, raiseError=True)
            except Exception as e:
                _logger.debug('Failed to delete "%s": %s', path, e)

    @classmethod
    def deleteAll(cls):
        for obj in cls.Instances.values():
            obj.deleteFile()

    @classmethod
    def remove(cls, filename, raiseError=True):
        """
        Remove a temporary file and delete the TempFile instance from the dict.

        :param filename: (str) the name of a temp file created by this class
        :param raiseError: (bool) if True, raise an exception if the filename is
            not a known TempFile.
        :return: none
        :raises PygcamException: if the path is not related to a TempFile instance.
        """
        try:
            obj = cls.Instances[filename]
            obj.deleteFile()
        except KeyError:
            if raiseError:
                raise PygcamException('No TempFile instance with name "%s"' % filename)




'''
.. The "gt" (gcamtool) commandline program

.. Copyright (c) 2017 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
import os
from pygcam.config import getParam, getParamAsInt
from pygcam.utils import mkdirs
from pygcam.project import Project
from .error import PygcamMcsUserError

# def _getJobNum():
#     import re
#
#     batchSystem = getParam('MCS.BatchSystem')
#     job_id_var = getParam("%s.%s" % (batchSystem, 'JOB_ID_VAR'))
#     jobIdStr = os.getenv(job_id_var, '')
#
#     result = re.search('\d+', jobIdStr)
#     jobNum = int(result.group(0)) if result else os.getpid()
#     return jobNum


def _dirFromNumber(n, prefix="", create=False):
    '''
    Compute a directory name using a 2-level directory structure that
    allows 1000 nodes at each level, accommodating up to 1 million files
    (0 to 999,999) in two levels.
    '''
    from numpy import log10     # lazy import

    maxnodes = getParamAsInt('MCS.MaxSimDirs') or 1000

    # Require a power of 10
    log = log10(maxnodes)
    if log != int(log):
        raise PygcamMcsUserError("MaxSimDirs must be a power of 10 (default value is 1000)")
    log = int(log)

    level1 = n / maxnodes
    level2 = n % maxnodes

    directory = os.path.join(prefix, str(level1).zfill(log), str(level2).zfill(log))
    if create:
        mkdirs(directory)

    return directory


def getSimDir(simId, create=False):
    '''
    Return and optionally create the path to the top-level simulation
    directory for the given simulation number, based on the SimsDir
    parameter specified in the config file.
    '''
    simsDir = getParam('MCS.RunSimsDir')
    if not simsDir:
        raise PygcamMcsUserError("Missing required config parameter 'RunSimsDir'")

    simDir = os.path.join(simsDir, 's%03d' % simId)  # name is of format ".../s001/"
    if create:
        mkdirs(simDir)

    return simDir


class Context(object):

    __slots__ = ['runId', 'simId', 'trialNum', 'scenario',
                 'baseline', 'groupName', 'projectName',
                 'useGroupDir', 'status']

    instances = {}      # Context instances keyed by runId

    def __init__(self, runId=None, projectName=None, simId=None, trialNum=None,
                 scenario=None, baseline=None, groupName=None, status=None,
                 store=True):
        self.runId     = runId
        self.simId     = simId
        self.trialNum  = trialNum
        self.scenario  = scenario
        self.baseline  = baseline
        self.status    = status

        self.projectName = projectName = projectName or getParam('GCAM.DefaultProject')
        project = Project.readProjectFile(projectName)

        self.groupName = groupName or project.scenarioSetup.defaultGroup
        self.useGroupDir = project.scenarioGroup.useGroupDir

        if store and runId:
            self.saveRunInfo()

    def saveRunInfo(self):
        Context.instances[self.runId] = self
        return self

    @classmethod
    def getRunInfo(cls, runId):
        return cls.instances.get(runId, None)

    def __str__(self):
        idTail = str(id(self))[-6:] # show last 6 digits only; enough to distinguish objs

        return "<Context id=%s prj=%s scn=%s grp=%s use=%s sim=%s trl=%s run=%s sta=%s>" % \
               (idTail, self.projectName, self.scenario, self.groupName, self.useGroupDir,
                self.simId, self.trialNum, self.runId, self.status)

    def setVars(self, projectName=None, simId=None, trialNum=None, scenario=None,
                baseline=None, groupName=None, status=None):
        '''
        Set elements of a context structure for all args that are not None.
        '''
        if projectName:
            self.projectName = projectName

        if simId is not None:
            self.simId = int(simId)

        if trialNum is not None:
            self.trialNum = int(trialNum)

        if scenario:
            self.scenario = scenario

        if baseline:
            self.baseline = baseline

        if groupName:
            self.groupName = groupName

        if status:
            self.status = status

    def getSimDir(self, create=False):
        '''
        Return and optionally create the path to the directory for a given sim.
        '''
        return getSimDir(self.simId, create=create)

    def getTrialDir(self, create=False):
        '''
        Return and optionally create the path to the directory for a given trial.
        '''
        simDir = getSimDir(self.simId, create=False)
        trialDir = _dirFromNumber(self.trialNum, prefix=simDir, create=create)
        return trialDir

    def getScenarioDir(self, create=False):
        '''
        Return and optionally create the path to the directory for a given experiment.
        '''
        trialDir = self.getTrialDir(create=False)
        scenarioDir = os.path.join(trialDir, self.scenario)
        if create:
            mkdirs(scenarioDir)

        return scenarioDir

    def getQueryResultsDir(self):
        trialDir = self.getTrialDir()
        result = os.path.join(trialDir, self.scenario, 'queryResults')
        return result

    def getDiffsDir(self):
        trialDir = self.getTrialDir()
        result = os.path.join(trialDir, self.scenario, 'diffs')
        return result

    @property
    def groupDir(self):
        return self.groupName if self.useGroupDir else ''

if __name__ == '__main__':
    c = Context()
    print(c)



#!/usr/bin/env python
#
# Feed trials to SLURM in managable batches and wait until these are
# underway before submitting more.
#

# TBD: Build this functionality into runsim?

import argparse
import subprocess
import sys
from time import sleep
from pygcam.config import getParam

def parseArgs():
    defaultSimId = 1
    defaultTrialCount = 100
    defaultMinutes = 5
    defaultMaxQueueSize = 400
    defaultQueueSleep = 60

    dfltProject = getParam('GCAM.DefaultProject')

    parser = argparse.ArgumentParser(description='''Run typical MCS analysis''')

    parser.add_argument('-c', '--trialsToQueue', type=int, default=defaultTrialCount,
                        help='''The number of trials to queue at once. Default is %d.''' % defaultTrialCount)

    parser.add_argument('-f', '--firstTrial', type=int, default=0,
                        help='''The trialNum of the first trial to process. Default is 0.''')

    parser.add_argument('-g', '--groupDir',
                        help='''The name of a scenario group to use as a directory name
                        between the project name and the scenario name.''')

    parser.add_argument('-P', '--project', default=dfltProject,
                        help='''The name of the project to run (current default project is "%s"''' % dfltProject)

    parser.add_argument('-t', '--trialsToRun', type=int, default=None,
                        help='''The total number of trials to create and run. If not provided, all trials
                        from the `firstTrial` to the last defined for the simulation will be run.''')

    parser.add_argument('-m', '--minutesToSleep', type=int, default=defaultMinutes,
                        help='''The number of minutes to sleep between tranches. Default is %d''' % defaultMinutes)

    parser.add_argument('-n', '--noRun', action='store_true',
                        help='''Print commands that would be executed, but don't run them.''')

    parser.add_argument('-q', '--maxQueueSize', type=int, default=defaultMaxQueueSize,
                        help='''Maximum number of queue slots to occupy. Default is %d.''' % defaultMaxQueueSize)

    parser.add_argument('-Q', '--queueSleep', type=int, default=defaultQueueSleep,
                        help='The number of seconds to sleep waiting for the queue to clear. Default is %d' % defaultQueueSleep)

    parser.add_argument('-s', '--simId', type=int, default=defaultSimId,
                        help='The id of the simulation to run. Default is %s.' % defaultSimId)

    parser.add_argument('-S', '--scenario', required=True,
                        help='''Comma-delimited list of scenarios to run''')

    args = parser.parse_args()
    return args

def run(cmd, ignoreError=False, printOnly=False):
    print(cmd)
    if printOnly:
        return

    status = subprocess.call(cmd, shell=True)

    if not ignoreError and status != 0:
        sys.exit(status)

# TBD: add ability to re-run trials of a given status, as runsim does with -r aborted,
# TBD: but using this "go slow" approach.
def main():
    from pygcammcs.Database import getDatabase

    args = parseArgs()

    project   = args.project
    simId     = args.simId
    scenarios = args.scenario.split(',')
    first     = args.firstTrial
    trials    = args.trials or getDatabase().getTrialCount(simId) - first
    noRun     = args.noRun
    count     = args.count

    last = first + trials - 1
    maxQueueSize = args.maxQueueSize
    delayMinutes = args.minutes
    delay = delayMinutes * 60
    queueSleep = args.queueSleep

    groupFlag = '-g ' + args.groupDir if args.groupDir else ''
    queueCmd = 'squeue -u $USER -p shared,slurm,short -h|wc -l'

    firstPass = True

    for scenario in scenarios:
        cmd = 'gt -P %s runsim -s %d -e %s %s' % (project, simId, scenario, groupFlag)

        for start in range(first, last, count):
            if firstPass:
                firstPass = False
            else:
                if noRun:
                    print('sleep(%d)' % delay)
                else:
                    sleep(delay)

            while True:
                result = subprocess.check_output(queueCmd, shell=True)
                queueCount = int(result)
                if queueCount + count > maxQueueSize:
                    print("Waiting %d sec for queue to clear. (%d entries)" % (queueSleep, queueCount))
                    sleep(queueSleep)
                else:
                    break

            trialSpec = '-t %d-%d' % (start, start + count - 1)
            run(cmd + trialSpec, printOnly=args.noRun)

main()



"""
.. Support for generating land-protection scenarios described in an XML file.

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from ..subcommand import SubcommandABC
from ..constants import UnmanagedLandClasses

DefaultTemplate = 'prot_{fraction}_{filename}'


class ProtectLandCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {#'aliases' : ['pro'],
                 'help': '''Generate versions of GCAM's land_input XML files that protect a
                 given fraction of land of the given land types in the given regions.'''}

        super(ProtectLandCommand, self).__init__('protect', subparsers, kwargs, group='project')

    def addArgs(self, parser):
        parser.add_argument('-b', '--backup', action='store_true',
                            help='''Make a copy of the output file, if it exists (with an added ~ after
                            filename) before writing new output. This option is ignored if a scenario
                            file is specified.''')

        parser.add_argument('-f', '--fraction', type=float, default=None,
                            help='''The fraction of land in the given land classes to protect. (Required,
                            unless a scenario file is specified, in which case this option is ignored.)''')

        parser.add_argument('--inPlace', action='store_true',
                            help='''Edit the file in place. This must be given explicitly, to avoid overwriting
                            files by mistake.''')

        parser.add_argument('-l', '--landClasses', action='append',
                            help='''The land class or classes to protect in the given regions. Multiple,
                            comma-delimited land types can be given in a single argument, or the -l flag can
                            be repeated to indicate additional land classes. By default, all unmanaged land
                            classes are protected. Allowed land classes are %s.
                            This option is ignored if a scenario file is specified. ''' % UnmanagedLandClasses)

        parser.add_argument('-m', '--mkdir', action='store_true',
                            help='''Make the output dir if necessary.''')

        parser.add_argument('-o', '--outDir', type=str, default='.',
                            help='''The directory into which to write the modified files.
                                    Default is current directory.''')

        parser.add_argument('-O', '--otherArable', action='store_true',
                            help='''Include OtherArableLand in the list of default land classes to protect.
                            This flag is ignored if the -l (--landClasses) argument is used.''')

        parser.add_argument('-t', '--template', type=str, default=DefaultTemplate,
                            help='''Specify a template to use for output filenames. The keywords {fraction}, {filename},
                            {regions}, and {classes} (with surrounding curly braces) are replaced by the following values
                            and used to form the name of the output files, written to the given output directory.
                            fraction: 100 times the given fraction (i.e., int(fraction * 100));
                            filename: the name of the input file being processed (e.g., land_input_2.xml or land_input_3.xml);
                            basename: the portion of the input filename prior to the extension (i.e., before '.xml');
                            regions: the given regions, separated by '-', or the word 'global' if no regions are specified;
                            classes: the given land classes, separated by '-', or the word 'unmanaged' if no land classes
                            are specified. The default pattern is "%s".
                            This option is ignored if a scenario file is specified.''' % DefaultTemplate)

        parser.add_argument('-r', '--regions', action='append',
                            help='''The region or regions for which to protect land. Multiple, comma-delimited
                            regions can be given in a single argument, or the -r flag can be repeated to indicate
                            additional regions. By default, all regions are protected.
                            This option is ignored if a scenario file is specified.''')

        parser.add_argument('-s', '--scenario', default=None,
                            help='''The name of a land-protection scenario defined in the file given by the --scenarioFile
                            argument or it's default value.''')

        parser.add_argument('-S', '--scenarioFile', default=None,
                            help='''An XML file defining land-protection scenarios. Default is the value
                            of configuration file parameter GCAM.LandProtectionXmlFile.''')

        parser.add_argument('-w', '--workspace', type=str, default=None,
                            help='''Specify the path to the GCAM workspace to use. The files in
                            {workspace}/input/gcam-data-system/xml/aglu-xml/land_input_{2,3}.xml
                            are used as inputs. Default is value of configuration parameter
                            GCAM.RefWorkspace.''')

        return parser

    def run(self, args, tool):
        from ..landProtection import protectLandMain

        protectLandMain(args)



#! /usr/bin/env python
'''
Created on 4/26/15

@author: rjp
'''
from ..subcommand import SubcommandABC

class BioConstraintsCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Bioenergy constraint generator'''}

        super(BioConstraintsCommand, self).__init__('bioConstraint', subparsers, kwargs)

    def addArgs(self, parser):
        from ..constraints import PolicyChoices, DefaultCellulosicCoefficients, DefaultYears

        parser.add_argument('-b', '--baseline', default=None,
                            help='The baseline on which the policy scenario is based')

        parser.add_argument('-B', '--biomassPolicyType', default='tax', choices=PolicyChoices,
                             help='Regional biomass policy type')

        parser.add_argument('-c', '--coefficients',
                            help='''A comma-separated string of year:coefficient pairs. This
                            sets the cellulosic ethanol conversion coefficients. Defaults to
                            standard GCAM values: %s.''' % DefaultCellulosicCoefficients)

        parser.add_argument('-e', '--cellEtohPolicyType', default='tax', choices=PolicyChoices,
                             help='Cellulosic ethanol policy type')

        parser.add_argument('-l', '--defaultLevel', type=float, default=0.0,
                            help='''Target cellulosic biofuel level (EJ). All or individual years
                            values can be set (overriding -l flag values) using the -L flag.''')

        parser.add_argument('-L', '--annualLevels', required=False,
                            help='''Optional cellulosic biofuel production levels by year. Value
                            must be a comma-delimited string of year:level pairs, where level is
                            given in EJ. If -l is not used to set default for all years, you must
                            specify values for all years using this option.''')

        # parser.add_argument('-m', '--fromMCS', action='store_true',
        #                      help="Used when calling from gcammcs so correct pathnames are computed.")

        parser.add_argument('-p', '--policy',
                            help='The policy scenario name')

        parser.add_argument('-P', '--purposeGrownPolicyType', default='subsidy', choices=PolicyChoices,
                             help='Purpose-grown biomass policy type')

        parser.add_argument('-R', '--resultsDir',
                            help='The parent directory holding the GCAM output workspaces')

        parser.add_argument('-s', '--switchgrass', action='store_true',
                             help="Generate constraints for switchgrass")

        parser.add_argument('-S', '--subdir', default='',
                             help='Sub-directory for local-xml files, if any')

        parser.add_argument('-x', '--xmlOutputDir',
                             help='''The directory into which to generate XML files. Defaults to
                             policy name in the current directory.''')

        parser.add_argument('-y', '--years', default=DefaultYears,
                            help='''Years to generate constraints for. Must be of the form
                            XXXX-YYYY. Default is "%s"''' % DefaultYears)
        return parser


    def run(self, args, tool):
        from ..constraints import genBioConstraints

        genBioConstraints(**vars(args))



# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.
import os
from .McsSubcommandABC import McsSubcommandABC
from pygcam.error import CommandlineError
from pygcam.log import getLogger

_logger = getLogger(__name__)

#
# Changes to make to boilerplate ipython parallel config files. Format is
# {filename : ((match, replacement), (match, replacement)), filename: ...}
#
FileChanges = {
    'ipcluster_config.py': (
        ("#c.IPClusterEngines.engine_launcher_class = 'LocalEngineSetLauncher",
         "c.IPClusterEngines.engine_launcher_class = '{scheduler}'"),

        ("#c.IPClusterEngines.n = ",    # not sure if number is constant across platforms...
         "# Attempt to allocate ~8 GB per engine\nc.IPClusterEngines.n = {engines}"),

        ("#c.IPClusterStart.delay = 1.0",
         "# No need to delay queuing engines with controller on login node\nc.IPClusterStart.delay = 0.1"),

        ("#c.SlurmLauncher.account = u''",
         "c.SlurmLauncher.account = u'{account}'"),

        ("#c.SlurmLauncher.timelimit = u''",
         "c.SlurmLauncher.timelimit = u'{walltime}'"),

        ("#c.SlurmControllerLauncher.batch_file_name = u'slurm_controller.sbatch'",
         "c.SlurmControllerLauncher.batch_template_file = u'slurm_mcs_controller.template'"),

        ("#c.SlurmEngineSetLauncher.batch_file_name = u'slurm_engine.sbatch'",
         "c.SlurmEngineSetLauncher.batch_template_file = u'slurm_mcs_engine.template'"),

        ("#c.PBSControllerLauncher.batch_file_name = u'pbs_controller'",
         "c.PBSControllerLauncher.batch_template_file = u'pbs_mcs_controller.template'"),

        ("#c.PBSEngineSetLauncher.batch_file_name = u'pbs_engine'",
         "c.PBSEngineSetLauncher.batch_template_file = u'pbs_mcs_engine.template'"),

        ("#c.LSFControllerLauncher.batch_file_name = u'lsf_controller'",
         "c.LSFControllerLauncher.batch_template_file = u'lsf_mcs_controller.template'"),

        ("#c.LSFEngineSetLauncher.batch_file_name = u'lsf_engine'",
         "c.LSFEngineSetLauncher.batch_template_file = u'lsf_mcs_engine.template'"),
    ),

    'ipcontroller_config.py': (
        ("#c.HubFactory.client_ip = u''",
         "c.HubFactory.client_ip = u'*'"),

        ("#c.HubFactory.engine_ip = u''",
         "c.HubFactory.engine_ip = u'*'"),
    )
}

class IppSetupCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Start an ipyparallel cluster after generating batch
        file templates based on parameters in .pygcam.cfg and the number of tasks to run.'''}
        super(IppSetupCommand, self).__init__('ippsetup', subparsers, kwargs)

    def addArgs(self, parser):
        defaultAccount = ''
        defaultProfile = 'pygcam'
        defaultEngines = 4
        defaultMinutes = 30
        schedulers = ('Slurm', 'PBS', 'LSF')
        defaultScheduler = schedulers[0]

        parser.add_argument('-a', '--account', default=defaultAccount,
                            help='''The account name to use to run jobs on the cluster system.
                            Used by Slurm only. Default is "%s"''' % defaultAccount)

        parser.add_argument('-e', '--engines', type=int, default=defaultEngines,
                            help='''Set default number of engines to allow per node.
                            This is overridden by runsim; this value is used when running
                            the cluster "manually". Default value is %d.''' % defaultEngines)

        parser.add_argument('-m', '--minutes', type=int, default=defaultMinutes,
                            help='''The default number of minutes to allocate
                            per GCAM run. (Used by Slurm only.) This is used for 
                            the "+b / --batch" and "gt run -D" options only. The 
                            "runsim" sub-command uses the value in 
                            IPP.MinutesPerRun. Default value is %d.''' % defaultMinutes)

        parser.add_argument('-p', '--profile', type=str, default=defaultProfile,
                            help='''The name of the ipython profile to create. Set config 
                            variable IPP.Profile to the same value. Default is
                            "%s".''' % defaultProfile),

        parser.add_argument('-s', '--scheduler', choices=schedulers, default=defaultScheduler,
                            help='''The resource manager / scheduler your system uses. 
                            Default is %s.''' % defaultScheduler)

        return parser   # for auto-doc generation


    def run(self, args, tool):
        from IPython.paths import get_ipython_dir
        import subprocess as subp

        minutes = args.minutes
        walltime = '%d:%02d:00' % (minutes/60, minutes%60)

        formatDict = {'account'  : args.account,
                      'engines'  : args.engines,
                      'scheduler': args.scheduler,
                      'minutes'  : minutes,
                      'walltime' : walltime}

        profile = args.profile

        ipythonDir = get_ipython_dir()
        profileDir = os.path.join(ipythonDir, 'profile_' + profile)

        if os.path.isdir(profileDir):
            raise CommandlineError('Ipython profile directory "%s" already exists. Delete it or choose another name.' % profileDir)

        cmd = 'ipython profile create --parallel ' + profile
        _logger.info('Running command: %s', cmd)
        subp.call(cmd, shell=True)

        for basename, tuples in FileChanges.items():
            pathname = os.path.join(profileDir, basename)
            backup   = pathname + '~'

            if not os.path.isfile(pathname):
                raise CommandlineError('Missing configuration file: "%s"' % pathname)

            _logger.info('Editing config file: %s', pathname)
            os.rename(pathname, backup)
            with open(backup, 'r') as input:
                lines = input.readlines()

            with open(pathname, 'w') as output:
                for line in lines:
                    for pattern, replacement in tuples:
                        if line.startswith(pattern):
                            line = replacement.format(**formatDict) + '\n'
                            break

                    output.write(line)



'''
.. Created on: 2/7/2017
   Fix for non-framework version of python on Mac OS X

.. Copyright (c) 2017 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
import matplotlib as mpl

_backend = 'Agg'

import platform
# if platform.system() == 'Darwin':
#     major, minor, rev = map(int, mpl.__version__.split('.'))
#     if major >= 2:
#         _backend = 'TkAgg'

mpl.use(_backend)

import matplotlib.pyplot as plt



from ..subcommand import SubcommandABC

class DiffCommand(SubcommandABC):
    def __init__(self, subparsers):
        helptext = 'Compute differences between CSV files generated by GCAM batch queries.'
        desc = '''

            '''
        kwargs = {'help' : helptext,
                  'description' : desc}
        super(DiffCommand, self).__init__('diff', subparsers, kwargs, group='project')

    def addArgs(self, parser):
        parser.add_argument('csvFiles', nargs='+',
                    help='''The files to process. For difference operations, the first file is treated
                    as the reference file whose time-series data is subtracted from that of each other
                    file. If missing, ".csv" suffixes are added to all arguments (the ".csv" is optional).''')

        parser.add_argument('-D', '--workingDir', default='.',
                            help='''The directory to change to before performing any operations''')

        parser.add_argument('-g', '--groupSum', default="",
                            help='''Group data for each timestep (or interpolated annual values) by the
                            given column, and sum all members of each group to produce a timeseries for
                            each group. Takes precedence over the simpler "-S" ("--sum") option.''')

        parser.add_argument('-i', '--interpolate', action="store_true",
                            help="Interpolate (linearly) annual values between timesteps.")

        parser.add_argument('-o', '--outFile', default='differences.csv',
                            help='''The name of the ".csv" or ".xlsx" file containing the differences
                            between each scenario and the reference. Default is "differences.csv".''')

        parser.add_argument('-c', '--convertOnly', default=False, action="store_true",
                            help='''Convert the given CSV files into an Excel workbook, one sheet per CSV file.''')

        parser.add_argument('-q', '--queryFile', default='',
                            help='''If the extension is ".xml" (case insensitive), the argument must be an XML file
                            holding a list of queries to run, with optional mappings specified to rewrite output.
                            This file has the same structure as the <queries> element in project.xml. If the file
                            doesn't end in ".xml", it must be a text file listing the names of queries to process,
                            one per line. NOTE: When --queryFile is specified, the two positional arguments are
                            required: the names of the baseline and policy scenarios, in that order.''')

        parser.add_argument('-r', '--rewriteSetsFile',
                            help='''An XML file defining query maps by name (default taken from
                            config parameter "GCAM.RewriteSetsFile")''')

        parser.add_argument('-S', '--sum', default=False, action="store_true",
                            help='''Sum all timestep (or interpolated annual values) to produce a single time-series.''')

        parser.add_argument('-s', '--skiprows', type=int, default=1,
                            help='''The number of rows to skip. Default is 1, which works for GCAM batch query output.
                            Use -s0 for outFile.csv''')

        parser.add_argument('-y', '--years', default="",
                            help='''Takes a parameter of the form XXXX-YYYY, indicating start and end years of interest.
                            Other years are dropped (except for annual outputs.)''')

        parser.add_argument('-Y', '--startYear', type=int, default=0,
                            help='''The year at which to begin interpolation''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        from ..diff import diffMain
        diffMain(args)



'''
.. Created on: 2/26/15

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
import os
import re
import subprocess

from .config import getParam, getParamAsBoolean, getParamAsFloat
from .error import ProgramExecutionError, GcamError, GcamSolverError, PygcamException
from .log import getLogger
from .scenarioSetup import createSandbox
from .utils import writeXmldbDriverProperties, getExeDir, pushd, parse_version_info
from .windows import IsWindows

_logger = getLogger(__name__)

PROGRAM = os.path.basename(__file__)

def setJavaPath(exeDir):
    '''
    Update the PATH to be able to find the Java dlls.
    Modeled on run-gcam.bat in the GCAM distribution.
    '''
    if not IsWindows:
        return

    javaHome = os.environ.get('JAVA_HOME', None)

    if not javaHome:
        # Use WriteLocalBaseXDB (4.2) or XMLDBDriver (>= 4.3) to print the java.home property
        # of the Java Runtime used to run it. Note if the runtime is not 64-bit it will only
        # print an error.
        with pushd(exeDir):
            versionInfo = parse_version_info()
            if versionInfo > (4, 2):
                classpath = getParam('GCAM.MI.ClassPath')
                command = 'java -cp "%s" XMLDBDriver --print-java-home' % classpath
            else:
                command = 'java WriteLocalBaseXDB'

            try:
                output = subprocess.check_output(str(command), shell=True)
            except Exception as e:
                raise PygcamException("Cannot get java home dir: %s" % e)

        os.environ['JAVA_HOME'] = javaHome = output and output.strip()

        if not javaHome:
            raise PygcamException("JAVA_HOME not set and failed to read java home directory from WriteLocalBaseXDB")

    if not os.path.isdir(javaHome):
        raise PygcamException('Java home (%s) is not a directory' % javaHome)

    # Update the PATH to be able to find the Java dlls
    # SET PATH=%PATH%;%JAVA_HOME%\bin;%JAVA_HOME%\bin\server"
    javaBin = os.path.join(javaHome, 'bin')
    javaBinServer = os.path.join(javaBin, 'server')
    envPath = os.environ.get('PATH', '')
    os.environ['PATH'] = path = envPath + ';' + javaBin + ';' + javaBinServer
    _logger.debug('PATH=%s', path)

    envClasspath = os.environ.get('CLASSPATH', '')
    envClasspath = ".;" + envClasspath if envClasspath else "."

    miClasspath = getParam('GCAM.MI.ClassPath')
    os.environ['CLASSPATH'] = classpath = envClasspath + ';' + javaBinServer + ';' + miClasspath
    _logger.debug('CLASSPATH=%s', classpath)

def _gcamWrapper(args):
    try:
        _logger.debug('Starting gcam with wrapper')
        gcamProc = subprocess.Popen(args, bufsize=0, stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT, close_fds=True)

    except Exception as e:
        msg = 'gcamWrapper failed to run command: {} ({})'.format(' '.join(args), e)
        raise PygcamException(msg)

    modelDidNotSolve = 'Model did not solve'
    pattern = re.compile('(.*(BaseXException|%s).*)' % modelDidNotSolve)

    gcamOut = gcamProc.stdout
    while True:
        line = gcamOut.readline()
        if line == '':
            break

        _logger.info(line.rstrip())          # see if this ends up in worker.log

        match = re.search(pattern, line)
        if match:
            gcamProc.terminate()
            msg = 'GCAM error: ' + match.group(0)
            if match.group(1) == modelDidNotSolve:
                raise GcamSolverError(msg)
            else:
                raise GcamError(msg)

    _logger.debug('gcamWrapper found EOF. Waiting for GCAM to exit...')
    status = gcamProc.wait()
    _logger.debug('gcamWrapper: GCAM exited with status %s', status)
    return status

#
# Comment taken from exe/run-gcam.command:
#
# We need to find where the Java development kit is installed.
# This could be the Apple supplied version which was provided up
# to 1.6 however was dropped subsequently and instead users may
# have an Oracle provided JDK.  The each take slightly different
# approaches to where libraries live and how to reference them so
# we will have to try to detect the appropriate location.
#
def linkToMacJava():
    import subprocess as subp
    from .error import PygcamException

    cmd = '/usr/libexec/java_home'
    javaHome = ''
    try:
        javaHome = subp.check_output(cmd).strip()
    except Exception:
        pass

    if not javaHome:
        raise PygcamException('Could not find Java install location using "%s"' % cmd)

    # If javaHome contains "1.6", use the Apple supplied version of java 1.6
    libPath = 'lib-stub' if '1.6' in javaHome else javaHome + '/jre/lib/server'

    owd = os.getcwd()
    refWorkspace = getParam('GCAM.RefWorkspace')
    os.chdir(refWorkspace)

    try:
        # Create a symlink to satisfy @rpath searches
        linkName = 'libs/java/lib'
        if not os.path.islink(linkName):
            cmd = "ln -s %s %s" % (libPath, linkName)
            status = subp.call(cmd, shell=True)
            if status != 0:
                raise PygcamException('Failed to create link using "%s"' % cmd)
    finally:
        os.chdir(owd)

def runGCAM(scenario, workspace=None, refWorkspace=None, scenariosDir=None, groupDir='',
            configFile=None, forceCreate=False, noRun=False, noWrapper=False):
    """

    :param scenario: (str) the scenario to run
    :param workspace: (str) path to the workspace to run in, or None, in which
       case the model is run in {GCAM.SandboxDir}/{scenario} if scenario is given
       otherwise, the default scenario in the configuration.xml in the GCAM.RefWorkspace
       is run.
    :param refWorkspace: (str) a workspace to copy files from to create the sandbox,
       if the workspace is not given, or doesn't exist.
    :param scenariosDir: (str) the directory in which the config.xml file for the
       given scenario is found. Defaults to GCAM.ScenariosDir, if given, or "."
    :param groupDir: (str) the name of the scenario group if group sub-directories
       are to be used when computing the location of the scenario's config.xml.
    :param configFile: (str) if scenario is not given, the name of a configuration
       file to run. If scenario is given, this parameter is ignored.
    :param forceCreate: (bool) if True, recreate the sandbox even if it already exists.
    :param noRun: (bool) if True, don't run the model, just create the sandbox and
       display the command that would be executed.
    :param noWrapper: (bool) if True, don't run GCAM inside a "wrapper" that reads
        output and kills the model run as soon as an error is detected.
    :return: none
    :raises ProgramExecutionError: if GCAM exits with non-zero status
    """
    import platform
    from .utils import parse_version_info

    if platform.system() == 'Darwin':
        linkToMacJava()

    workspace = workspace or (os.path.join(getParam('GCAM.SandboxDir'), scenario)
                                       if scenario else getParam('GCAM.RefWorkspace'))

    if not os.path.lexists(workspace) or forceCreate:
        createSandbox(workspace, srcWorkspace=refWorkspace, forceCreate=forceCreate)

    exeDir = getExeDir(workspace, chdir=True)
    setJavaPath(exeDir)     # required for Windows; a no-op otherwise

    version = parse_version_info()

    # These features didn't exist in version 4.2
    if version > (4, 2) and not (getParamAsBoolean('GCAM.RunQueriesInGCAM') or
                                 getParamAsBoolean('GCAM.InMemoryDatabase')):    # this implies RunQueriesInGCAM
        # Write a "no-op" XMLDBDriver.properties file
        writeXmldbDriverProperties(inMemory=False, outputDir=exeDir)

    if scenario:
        # Translate scenario name into config file path, assuming that for scenario
        # FOO, the configuration file is {scenariosDir}/{groupDir}/FOO/config.xml
        scenariosDir = os.path.abspath(scenariosDir or getParam('GCAM.ScenariosDir') or '.')
        configFile   = os.path.join(scenariosDir, groupDir, scenario, "config.xml")
    else:
        configFile = os.path.abspath(configFile or os.path.join(exeDir, 'configuration.xml'))

    gcamPath = os.path.abspath(getParam('GCAM.Executable'))
    gcamArgs = [gcamPath, '-C%s' % configFile]  # N.B. GCAM (< 4.2) doesn't allow space between -C and filename

    command = ' '.join(gcamArgs)
    if noRun:
        print(command)
    else:
        _logger.info('Running: %s', command)

        noWrapper = IsWindows or noWrapper     # never use the wrapper on Windows
        exitCode = subprocess.call(gcamArgs, shell=False) if noWrapper else _gcamWrapper(gcamArgs)

        if exitCode != 0:
            raise ProgramExecutionError(command, exitCode)


def gcamMain(args):
    runGCAM(args.scenario, workspace=args.workspace, refWorkspace=args.refWorkspace,
            scenariosDir=args.scenariosDir, groupDir=args.groupDir, configFile=args.configFile,
            forceCreate=args.forceCreate, noRun=args.noRun, noWrapper=args.noWrapper)



# Copyright (c) 2012-2015. The Regents of the University of California (Regents)
# and Richard Plevin. See the file COPYRIGHT.txt for details.
import os
import numpy as np

from pygcam.matplotlibFix import plt

import pandas as pd
import seaborn as sns

from pygcam.config import getParam, getParamAsBoolean
from pygcam.log import getLogger
from pygcam.utils import mkdirs

from .error import PygcamMcsSystemError, PygcamMcsUserError
from .Database import getDatabase, Input

_logger = getLogger(__name__)

DEFAULT_BIN_COUNT = 3

def makePlotPath(value, simId):
    plotDir = getParam('MCS.PlotDir')
    subDir  = os.path.join(plotDir, "s%d" % simId)
    mkdirs(subDir)
    plotType = getParam('MCS.PlotType')
    path = os.path.join(subDir, "%s.%s" % (value, plotType))
    #print "Plot path: ", path
    return path

def printExtraText(fig, text, loc='top', color='lightgrey', weight='ultralight', fontsize='xx-small'):
    """
    Print 'extra' text at the top, bottom, right, or left edge of the figure.
    """
    if not text:
        return

    rot = 0
    ha  = 'center'
    va  = 'center'
    x   = 0.5
    y   = 0.5

    if loc == 'top':
        y = 0.98
        va = 'top'
    elif loc == 'bottom':
        y = 0.02
        va = 'bottom'
    elif loc == 'right':
        x = 0.98
        ha = 'right'
        rot = 270
    else: # left
        x = 0.02
        ha = 'left'
        rot = 90

    fig.text(x, y, text, color=color, weight=weight, fontsize=fontsize, va=va, ha=ha, rotation=rot)


def plotHistogram(values, xlabel=None, ylabel=None, title=None, xmin=None, xmax=None,
                  extra=None, extraColor='grey', extraLoc='right',
                  hist=True, showCI=False, showMean=False, showMedian=False,
                  color=None, shade=False, kde=True, show=True, filename=None):

    fig = plt.figure()

    style    = "white"
    colorSet = "Set1"
    sns.set_style(style)
    sns.set_palette(colorSet, desat=0.6)
    red, blue, green, purple = sns.color_palette(colorSet, n_colors=4)

    color = blue if color is None else color
    count = values.count()
    bins  = count / 10 if count > 150 else (count / 5 if count > 50 else (count / 2 if count > 20 else None))
    sns.distplot(values, hist=hist, bins=bins, kde=kde, color=color, kde_kws={'shade': shade})

    #sns.axlabel(xlabel=xlabel, ylabel=ylabel)
    if xlabel:
        plt.xlabel(xlabel) # , size='large')
    if ylabel:
        plt.ylabel(ylabel) # , size='large')

    sns.despine()

    if title:
        t = plt.title(title)
        t.set_y(1.02)

    printExtraText(fig, extra, color=extraColor, loc=extraLoc)

    if xmin is not None or xmax is not None:
        ax = plt.gca()
        ax.set_autoscale_on(False)
        ax.set_xlim(xmin, xmax)

    if showCI or showMean:
        ymin, ymax = plt.ylim()
        xmin, xmax = plt.xlim()
        textSize = 9
        labely   = ymax * 0.95
        deltax   = (xmax-xmin) * 0.01

        if showCI:
            color = red
            ciLow  = np.percentile(values, 2.5)
            ciHigh = np.percentile(values, 97.5)
            plt.axvline(ciLow,  color=color, linestyle='solid', linewidth=2)
            plt.axvline(ciHigh, color=color, linestyle='solid', linewidth=2)
            plt.text(ciLow  + deltax, labely, '2.5%%=%.2f'  % ciLow,  size=textSize, rotation=90, color=color)
            plt.text(ciHigh + deltax, labely, '97.5%%=%.2f' % ciHigh, size=textSize, rotation=90, color=color)

        if showMean:
            color = green
            mean = np.mean(values)
            plt.axvline(mean, color=color, linestyle='solid', linewidth=2)
            plt.text(mean + deltax, labely, 'mean=%.2f' % mean, color=color, size=textSize, rotation=90)

        if showMedian:
            color = purple
            median = np.percentile(values, 50)
            labely = ymax * 0.50
            plt.axvline(median, color=color, linestyle='solid', linewidth=2)
            plt.text(median + deltax, labely, 'median=%.2f' % median, color=color, size=textSize, rotation=90)

    if show:
        plt.show()

    if filename:
        _logger.info("plotHistogram writing to: %s", filename)
        fig.savefig(filename)

    plt.close(fig)


def plotTornado(data, colname='value', labelsize=9, title=None, color=None, height=0.8,
                limit=None, rlabels=None, xlabel='Contribution to variance', figsize=None,
                show=True, filename=None, extra=None, extraColor='grey', extraLoc='right'):
    '''
    :param data: A sorted DataFrame or Series indexed by variable name, with
                 column named 'value' and if rlabels is set, a column of that
                 name holding descriptive labels to display.
    :param labelsize: font size for labels
    :param title: If not None, the title to show
    :param color: The color of the horizontal bars
    :param height: Bar height
    :param limit: The maximum number of variables to display
    :param rlabels: If not None, the name of a column holding values to show on the right
    :param xlabel: Label for X-axis
    :param figsize: tuple for desired figure size. Defaults to (12,6) if rlabels else (8,6).
    :param show: If True, the figure is displayed on screen
    :param filename: If not None, the figure is saved to this file
    :param extra: Extra text to display in the lower left corner of the plot
    :return: nothing
    '''
    count, cols = data.shape

    if limit is None:
        limit = 20

    if 0 < limit < count:
        data = data[:limit]            # Truncate the DF to the top "limit" rows
        count = limit

    # Reverse the order so the larger (abs) values are at the top
    revIndex = list(reversed(data.index))
    data = data.loc[revIndex]

    itemNums = range(count)
    ypos = np.array(itemNums) - 0.08   # goose the values to better center labels

    if not figsize:
        figsize = (12, 6) if rlabels else (8, 6)

    #fig = plt.figure(figsize=figsize)
    #fig = plt.figure(facecolor='white', figsize=figsize)
    #plt.plot()

    # if it's a dataframe, we expect to find the data in the value column
    values = data if isinstance(data, pd.Series) else data[colname]

    if color is None:
        color = sns.color_palette("deep", 1)

    # TBD: This looks like it has been resolved; try this again using seaborn
    # tried pandas; most of the following manipulations can be handled in one call, but
    # it introduced an ugly dashed line at x=0 which I didn't see how to remove. Maybe
    # address this again if seaborn adds a horizontal bar chart.
    values.plot(kind="barh", color=sns.color_palette("deep", 1), figsize=figsize,
                xlim=(-1, 1), ylim=(-1, count), xticks=np.arange(-0.8, 1, 0.2))

    plt.xlabel(xlabel)

    right = 0.6 if rlabels else 0.9
    plt.subplots_adjust(left=0.3, bottom=0.1, right=right, top=0.9)  # more room for rlabels

    fig = plt.gcf()
    ax  = plt.gca()

    ax.xaxis.tick_top()
    ax.tick_params(axis='x', labelsize=labelsize)
    ax.tick_params(axis='y', labelsize=labelsize)
    ax.set_yticklabels(data.index)
    ax.set_yticks(itemNums)

    if rlabels:
        ax2 = plt.twinx()
        plt.ylim(-1, count)
        ax2.tick_params(axis='y', labelsize=labelsize)
        ax2.set_yticklabels(data[rlabels])
        ax2.set_yticks(itemNums)

        for t in ax2.xaxis.get_major_ticks() + ax2.yaxis.get_major_ticks():
            t.tick1On = False
            t.tick2On = False

    # show vertical grid lines only
    ax.yaxis.grid(False)
    ax.xaxis.grid(True)

    # Remove tickmarks from both axes
    for t in ax.xaxis.get_major_ticks() + ax.yaxis.get_major_ticks():
        t.tick1On = False
        t.tick2On = False

    if title:
        plt.title(title, y=1.05)  # move title up to avoid tick labels

    printExtraText(fig, extra, loc=extraLoc, color=extraColor)

    if show:
        plt.show()

    if filename:
        _logger.debug("Saving tornado plot to %s" % filename)
        fig.savefig(filename)

    plt.close(fig)


def plotConvergence(simId, expName, paramName, values, show=True, save=False):
    '''
    Examine the first 3 moments (mean, std, skewness) in the data set
    for increasing number (N) of values, growing by the given increment.
    Optionally plot the relationship between each of the moments and N,
    so we can when (if) convergence occurs.
    '''
    _logger.debug("Generating convergence plots...")
    count = values.count()
    results = {'Mean': [], 'Stdev': [], 'Skewness': [], '95% CI': []}

    increment = min(100, int(count/20))
    nValues = range(increment, count+increment-1, increment)

    for N in nValues:
        sublist = values[:N]
        results['Mean'].append(sublist.mean())
        results['Stdev'].append(sublist.std())
        results['Skewness'].append(sublist.skew())

        ciLow  = np.percentile(sublist, 2.5)
        ciHigh = np.percentile(sublist, 97.5)
        results['95% CI'].append(ciHigh - ciLow)

    # Insert zero value at position 0 for all lists to ensure proper scaling
    nValues.insert(0,0)
    for dataList in results.values():
        dataList.insert(0,0)

    labelsize=12
    for key, values in results.iteritems():
        plt.clf()   # clear previous figure
        ax = plt.gca()
        ax.tick_params(axis='x', labelsize=labelsize)
        ax.tick_params(axis='y', labelsize=labelsize)
        plt.plot(nValues, results[key])
        plt.title("%s" % paramName, size='large')
        ax.yaxis.grid(False)
        ax.xaxis.grid(True)
        plt.xlabel('Trials', size='large')
        plt.ylabel(key, size='large')
        plt.figtext(0.12, 0.02, "SimId=%d, Exp=%s" % (simId, expName),
                    color='black', weight='roman', size='x-small')

        if save:
            filename = makePlotPath("%s-s%02d-%s-%s" % (expName, simId, paramName, key), simId)
            _logger.debug("Saving convergence plot to %s" % filename)
            plt.savefig(filename)

        if show:
            plt.show()

    fig = plt.gcf()
    plt.close(fig)


# Could use series.describe() but I like this format better
def printStats(series):
    name   = series.name
    count  = series.count()
    mean   = series.mean()
    median = series.median()
    stdev  = series.std()
    skew   = series.skew()
    minv   = series.min()
    maxv   = series.max()
    ciLow  = series.quantile(0.025)
    ciHigh = series.quantile(0.975)
    ciLower  = series.quantile(0.01)
    ciHigher = series.quantile(0.99)

    print('''
%s:
     count: %d
      mean: %.2f
    median: %.2f
     stdev: %.2f
      skew: %.2f
       min: %.2f
       max: %.2f
    95%% CI: [%.2f, %.2f]
    99%% CI: [%.2f, %.2f]''' % (name, count, mean, median, stdev, skew, minv, maxv,
                                ciLow, ciHigh, ciLower, ciHigher))

def normalizeSeries(series):
    'Normalize a series by dividing each element by the sum'
    total = series.sum()
    return series / total

def normalizeDF(df):
    '''
    Perform max-min normalization on all columns.
    :param df: (pandas.DataFrame) data to operate on
    :return: (pandas.DataFrame) normalized values.
    '''
    dfMin = df.min()
    df = (df - dfMin) / (df.max() - dfMin)
    return df

def spearmanCorrelation(inputs, results):
    '''
    Compute Spearman ranked correlation and normalized Spearman ranked
    correlation between values in a DataFrame of inputs and a Series of
    results. Returns a Series with the spearman rank correlation values.

    :param inputs: (pandas.DataFrame) input values for each parameter and trial
    :param results: (pandas.Series) values for one model result, per trial
    :return: (pandas.Series) rank correlations of each input to the output vector.
    '''
    corrList = [results.corr(inputs[col], method='spearman') for col in inputs.columns]
    spearman = pd.Series(data=corrList, index=inputs.columns, name='spearman')

    return spearman


def plotSensitivityResults(varName, data, filename=None, extra=None, limit=None, printIt=True):
    '''
    Prints results and generates a tornado plot with normalized squares of Spearman
    rank correlations between an output variable and all input variables.
    '''
    # Sort descending by absolute value (normalized are all positive from squaring)
    data.sort(columns=['normalized'], ascending=False, inplace=True)

    if printIt:
        print("UNCERTAINTY IMPORTANCE (%s)" % varName)
        print("----------------------")
        print(data.to_string(columns=['spearman', 'value'], float_format="{:4.2f}".format))

    title = 'Sensitivity of %s' % varName
    plotTornado(data, title=title, show=False, filename=filename, extra=extra, limit=limit)

# Deprecated?
def plotGroupSensitivityResults(varName, data, filename=None, extra=None, limit=None, printIt=True):
    '''
    Sum the normalized contribution to variance for subscripted parameters,
    along with contribution from unsubscripted ones. For example, we sum
    the contributions for "ETA[1,5]" and "ETA[1,6]" into "ETA".
    '''
    totals = pd.Series(name='totals')

    for idx, row in data.iterrows():
        # Lop off bracketed indices at first '['.
        # e.g., from "ETA[1,5]", we'll extract "ETA"
        pos = idx.find('[')
        paramName = idx if pos < 0 else idx[0:pos]

        # Sum the values; initialize on demand
        try:
            totals[paramName] += row.value
        except KeyError:
            totals[paramName]  = row.value

    df = pd.DataFrame(totals)

    negatives = (totals < 0)
    df['sign'] = 1
    df.ix[negatives, 'sign'] = -1

    df['absval']    = totals * df['sign']
    df['normalize'] = normalizeSeries(df['absval'])
    df['value']     = df['normalize'] * df['sign']

    # Sort by absolute value
    df.sort_values('absval', ascending=False, inplace=True)

    db = getDatabase()
    with db.sessionScope() as session:
        result = session.query(Input.paramName, Input.description).distinct().all()

    resultDF = pd.DataFrame(result)
    resultDF.fillna('')
    df['description'] = resultDF['description']

    if printIt:
        print("\nPARAMETER GROUPS (%s)" % varName)
        print("----------------")
        print(df.to_string(columns=['value', 'description'], float_format="{:4.2f}".format))

    title = 'Sensitivity of %s' % varName
    plotTornado(df, title=title, figsize=None, show=False, filename=filename, limit=limit, extra=extra)


def plotInputDistributions(simId, inputDF):
    '''Plot the input values individually to test that the distributions are as expected'''

    showHist  = True
    showKDE   = False
    showShade = getParamAsBoolean('MCS.PlotShowShading')

    for heading, series in inputDF.iteritems():
        plotHistogram(series, showCI=False,
                      xlabel='Parameter value', ylabel='Probability density',
                      title='Distribution for values of %s' % heading,
                      color=None, hist=showHist, kde=showKDE, shade=showShade,
                      show=False, filename=makePlotPath(heading, simId))


def plotOutputDistribution(simId, expName, resultSeries, resultName, xlabel, trials):
    filename = makePlotPath('%s-s%02d-%s-%d-trials' % (expName, simId, resultName, resultSeries.count()), simId)

    showHist  = getParamAsBoolean('MCS.PlotShowHistogram')
    showKDE   = getParamAsBoolean('MCS.PlotShowKDE')
    showShade = getParamAsBoolean('MCS.PlotShowShading')

    numValues = resultSeries.count()

    plotHistogram(resultSeries, xlabel=xlabel, ylabel='Probability density',
                  title='Frequency distribution for %s' % resultName,
                  extra='SimId=%d, Exp=%s, Trials=%d/%d' % (simId, expName, numValues, trials),
                  color=None, hist=showHist, kde=showKDE, shade=showShade,
                  showCI=True, showMean=True, showMedian=True, show=False, filename=filename)

# TBD: If row/col are obsolete, this info can now be read from trialData.csv or data.sa
def readParameterValues(simId, trials):
    def makeKey(paramName, row, col):
        return paramName
        #return "%s[%d][%d]" % (paramName, row, col) if row or col else paramName

    db = getDatabase()

    paramTuples = db.getParameters()        # Returns paramName, row, col
    paramNames  = map(lambda tup: makeKey(*tup), paramTuples)   # names like "foo[1][14]"
    inputDF     = pd.DataFrame(index=xrange(trials), columns=paramNames, dtype=float)
    _logger.debug("Found %d distinct parameter names" % len(paramNames))

    paramValues = db.getParameterValues(simId, asDataFrame=False)
    numParams = len(paramValues)
    _logger.info('%d parameter values read' % numParams)

    for row, col, value, trialNum, pname in paramValues:
        key = makeKey(pname, row, col)
        inputDF[key][trialNum] = value

    return inputDF

def _fixColname(name):
    pos = name.find('[')
    return name[:pos] if pos >= 0 else name

def exportInputs(exportFile, inputs):
    df = inputs.copy()

    # remove [x][y] from colnames
    df.columns = map(_fixColname, df.columns)

    _logger.debug("Exporting data to '%s'", exportFile)
    df.to_csv(exportFile)

# TBD: export in format useful with SALib, at least as an option
def exportResults(simId, resultList, expList, exportFile, sep=','):
    db = getDatabase()
    df = None

    for expName in expList:
        for resultName in resultList:
            # resultDf has 'trialNum' as index, 'value' holds float value
            resultDf = db.getOutValues(simId, expName, resultName)
            if resultDf is None:
                raise PygcamMcsUserError('No results were found for sim %d, experiment %s, result %s' % (simId, expName, resultName))

            # Add columns needed for boxplots
            resultDf['expName'] = expName
            resultDf['resultName'] = resultName

            resultDf.rename(columns = {resultName:'value'}, inplace=True)

            if df is None:
                df = resultDf
            else:
                df = pd.concat([df, resultDf])

    _logger.debug("Exporting results to '%s'", exportFile)
    df.to_csv(exportFile, sep=sep)

#
# Based on ema_workbench/core/utils.py:save_results()
#
def saveForEMA(simId, expNames, resultNames, inputDF, filename):
    """
    Save simulation results to the specified tar.gz file. The results are
    stored as csv files. There is an x.csv, and a csv for each outcome. In
    addition, there is a metadata csv which contains the datatype information
    for each of the columns in the x array. Unlike the version of this function
    in the EMA Workbench, this version collects data from the SQL database to
    generate a file in the required format.

    :param simId: (int) the id of the simulation
    :param expNames: (list of str) the names of the experiments to save results for
    :param resultNames: (list of str) all model input values, each row holding values for 1 trial
    :param inputDF: (pandas.DataFrame) the input data
    :param filename: (str) the path of the file
    :raises: IOError if file not found
    :return: none
    """
    from io import BytesIO
    import tarfile
    import time

    def add_file(tgzfile, string_to_add, filename):
        tarinfo = tarfile.TarInfo(filename)
        tarinfo.size = len(string_to_add)
        tarinfo.mode = 0644
        tarinfo.mtime = time.time()
        tgzfile.addfile(tarinfo, BytesIO(string_to_add.encode('UTF-8')))

    db = getDatabase()

    # InValue.row, InValue.col, InValue.value, Trial.trialNum, Input.paramName
    rows = inputDF.shape[0]

    with tarfile.open(filename, 'w:gz') as z:
        # Write the input values to the zipfile
        expData = inputDF.to_csv(None, sep=',', index=False) # index_label='trialNum'
        add_file(z, expData, 'experiments.csv')

        # Write experiment metadata
        dtypes = inputDF.dtypes
        # list(dtypes.iteritems()) produces results like:
        # [('A', dtype('int64')), ('B', dtype('int64')), ('C', dtype('float64'))] and
        # map(lambda dt: (dt[0], dt[1].descr), dtypes.iteritems()) produces:
        # [('A', [('', '<i8')]), ('B', [('', '<i8')]), ('C', [('', '<f8')])]
        # So, map(lambda dt: (dt[0], dt[1].descr[0][1]), dtypes.iteritems()) produces:
        # [('A', '<i8'), ('B', '<i8'), ('C', '<f8')]
        tuples = map(lambda dt: (dt[0], dt[1].descr[0][1]), dtypes.iteritems())
        strings = ["{},{}".format(name, dtype) for name, dtype in tuples]
        fileText = "\n".join(strings) + '\n'
        add_file(z, fileText, 'experiments metadata.csv')

        # Write outcome metadata    # TBD: deal with timeseries

        # outcome_meta = ["{},{}".format(outcome, ','.join(outcomes[outcome].shape))
        #                 for outcome in resultNames]
        strings = ["{},{}".format(resultName, rows) for resultName in resultNames]
        fileText = "\n".join(strings) + '\n'
        add_file(z, fileText, "outcomes metadata.csv")

        # Write outcomes
        for expName in expNames:
            for resultName in resultNames:
                outValueDF = db.getOutValues(simId, expName, resultName) # cols are trialNum and value; might need to do outValueDF[resultName].to_csv
                allTrialsDF = pd.DataFrame(index=xrange(rows))           # ensure that all trials are represented (with NA if need be)
                allTrialsDF[resultName] = outValueDF[resultName]
                fileText = allTrialsDF.to_csv(None, header=False, index=False)
                fname = "{}-{}.csv".format(resultName, expName)
                add_file(z, fileText, fname)

    print("Results saved successfully to {}".format(filename))

def getCorrDF(inputs, output):
    '''
    Generate a DataFrame with rank correlations between each input vector
    and the given output vector, and sort by abs(correlation), descending.

    :param inputs: (pandas.DataFrame) input values for each parameter and trial
    :param output: (pandas.Series) output values for one result, per trial
    :return: (pandas.DataFrame) two columns, "spearman" and "abs", the prior
       holding the Spearman correlations between each input and the output
       vector, and the latter with the absolute values of these correlations.
       The DataFrame is indexed by variable name and sorted by "abs", descending.
    '''
    corrDF = pd.DataFrame(spearmanCorrelation(inputs, output))
    corrDF['abs'] = corrDF.spearman.abs()
    corrDF.sort_values('abs', ascending=False, inplace=True)
    return corrDF

def binColumns(inputDF, bins=DEFAULT_BIN_COUNT):
    columns = inputDF.columns
    binned = pd.DataFrame(columns=columns)
    for col in columns:
        s = inputDF[col]
        binned[col] = pd.cut(s, bins, labels=False)

    return binned

# TBD: Finish refactoring this
class Analysis(object):
    def __init__(self, simId, scenarioNames, resultNames, limit=0):
        self.simId = simId
        self.scenarioNames = scenarioNames
        self.resultNames = resultNames
        self.limit = limit

        self.db = getDatabase()
        self.trials = self.db.getTrialCount(simId) if limit <= 0 else limit

        if not self.trials:
            raise PygcamMcsUserError('No trials were found for simId %d' % simId)

        self.inputDF = None
        self.resultDict = {}   # DFs of results for a scenario, keyed by scenario name

    def getInputs(self):
        '''
        Read inputs for the given simId. If already read, return the
        cached DataFrame.

        :return: (pandas.DataFrame) input values for all parameters
        '''
        if self.inputDF is not None:
            return self.inputDF

        self.inputDF = df = readParameterValues(self.simId, self.trials)
        # TBD: [x][y] is probably obsolete
        # remove [x][y] subscripts from colnames
        df.columns = map(_fixColname, df.columns)
        return df

    def exportInputs(self, exportFile, columns=None, sep=','):
        '''
        Export the inputs for the current simulation / scenario. If
        provided, limit the set of inputs to the named columns.

        :param columns: (iterable of str) names of columns to export
        :return: none
        '''
        df = self.getInputs()
        if columns:
            df = df[columns]

        _logger.debug("Exporting data to '%s'", exportFile)
        df.to_csv(exportFile, sep=sep)

    def getResults(self, scenarioList=None, resultList=None):
        '''
        Get the results for the given result names, or, if none are
        specified, for the results identified when at instantiation.
        Results are cached and thus read only once from the database.

        :param scenarioList: (iterable of str) the scenarios for which
           to get results
        :param resultList: (iterable of str) the results to export
        :param sep: (str) column separator to use in output file
        :return: none
        '''
        db = self.db
        simId = self.simId
        resultDict = self.resultDict

        resultList = resultList or self.resultNames
        scenarioList = scenarioList or self.scenarioNames

        for scenario in scenarioList:
            resultDF = resultDict.get(scenario)

            for resultName in resultList:
                # returns DF with 'trialNum' as index, 'value' holds float value
                if resultDF is None or resultName not in resultDF.columns:
                    values = db.getOutValues(simId, scenario, resultName, limit=self.limit)
                    if values is None:
                        raise PygcamMcsUserError(
                            'No results were found for sim %d, experiment %s, result %s' % (simId, scenario, resultName))

                    resultDF = values if not resultDF else pd.concat([resultDF, values])

            resultDict[scenario] = resultDF

        return resultDict

    def exportResults(self, exportFile, scenarioList=None, resultList=None, sep=','):
        '''
        Export the results for the given scenario and result names, or, if not
        specified, the ones identified when at instantiation.

        :param exportFile: (str) filename to create
        :param scenList: (iterable of str) the scenarios for which to export results
        :param resultList: (iterable of str) the results to export
        :param sep: (str) column separator to use in output file
        :return: none
        '''
        resultDict = self.getResults(scenarioList=scenarioList, resultList=resultList)
        exportDF = None

        for scenario in scenarioList:
            resultDF = resultDict[scenario]

            for resultName in resultList:
                # The resultDF has 'trialNum' as index, 'value' holds float value.
                # Denormalize these to store all scenarios' results in one DF.
                df = resultDF[resultName].copy()
                df.rename(columns={resultName: 'value'}, inplace=True)

                # Add columns needed for boxplots
                df['expName'] = scenario
                df['resultName'] = resultName

                # concatenate each result set below the accumulated data
                exportDF = pd.concat([exportDF, df]) if exportDF else df

        _logger.debug("Exporting results to '%s'", exportFile)
        df.to_csv(exportFile, sep=sep)

    def plotInputDistributions(self):
        '''Plot the input values individually to test that the distributions are as expected'''

        showHist = True
        showKDE = False
        showShade = getParamAsBoolean('MCS.PlotShowShading')
        inputDF = self.getInputs()
        simId = self.simId

        for heading, series in inputDF.iteritems():
            plotHistogram(series, showCI=False,
                          xlabel='Parameter value', ylabel='Probability density',
                          title='Distribution for values of %s' % heading,
                          color=None, hist=showHist, kde=showKDE, shade=showShade,
                          show=False, filename=makePlotPath(heading, simId))

    def plotUncertaintyImportance(self, inputDF, resultSeries, filename=None,
                                  extra=None, printIt=True):
        '''
        Prints results and generates a tornado plot with normalized squares of Spearman
        rank correlations between an output variable and all input variables.
        '''
        spearman = spearmanCorrelation(inputDF, resultSeries)
        data = pd.DataFrame(spearman)
        squared = spearman ** 2
        data['normalized'] = squared / squared.sum()
        data['sign'] = 1
        data.ix[(data.spearman < 0), 'sign'] = -1
        data['value'] = data.normalized * data.sign     # normalized squares with signs restored

        # Sort descending by normalized values (all are positive from squaring)
        data.sort(columns=['normalized'], ascending=False, inplace=True)

        varName = resultSeries.name

        if printIt:
            print("UNCERTAINTY IMPORTANCE (%s)" % varName)
            print("----------------------")
            print(data.to_string(columns=['spearman', 'value'], float_format="{:4.2f}".format))

        title = 'Sensitivity of %s' % varName
        plotTornado(data, title=title, show=False, filename=filename, extra=extra)

    def plotParallelCoordinates(self, inputDF, resultSeries, numInputs=None,
                                filename=None, extra=None, inputBins=None,
                                outputLabels=['Low', 'Medium', 'High'],
                                quantiles=False, normalize=True, invert=False,
                                show=False, title=None, rotation=None):
        '''
        Plot a parallel coordinates figure.

        :param inputDF: (pandas.DataFrame) trial inputs
        :param resultSeries: (pandas.Series) results to categorize lines
        :param numInputs: (int) the number of inputs to plot, choosing these
           from the most-highly correlated (or anti-correlated) to the lowest.
           If not provided, all variables in `inputDF` are plotted.
        :param filename: (str) name of graphic file to create
        :param extra: (str) text to draw down the right side, labeling the figure
        :param inputBins: (int) the number of bins to use to quantize inputs
        :param quantiles: (bool) create bins with equal numbers of values rather than
           bins of equal boundary widths. (In pandas terms, use qcut rather than cut.)
        :param normalize: (bool) normalize values to percentages of the range for each var.
        :param invert: (bool) Plot negatively correlated values as (1 - x) rather than (x).
        :param outputLabels: (list of str) labels to assign to outputs (and thus the number
           of bins to group the outputs into.)
        :param title: (str) Figure title
        :param show: (bool) If True, show the figure.
        :return: none
        '''
        from pandas.tools.plotting import parallel_coordinates

        corrDF = getCorrDF(inputDF, resultSeries)
        numInputs = numInputs or len(corrDF)
        cols = list(corrDF.index[:numInputs])

        # isolate the top-correlated columns
        inputDF = inputDF[cols]

        # trim down to trials with result (in case of failures)
        inputDF = inputDF.ix[resultSeries.index]

        if normalize or invert:
            inputDF = normalizeDF(inputDF)

        if invert:
            for name in cols:
                # flip neg. correlated values to reduce line crossings
                if corrDF.spearman[name] < 0:
                    inputDF[name] = 1 - inputDF[name]
                    inputDF.rename(columns={name: "(1 - %s)" % name}, inplace=True)
            cols = inputDF.columns

        # optionally quantize inputs into the given number of bins
        plotDF = binColumns(inputDF, bins=inputBins) if inputBins else inputDF.copy()

        # split results into equal-size or equal-quantile bins
        cutFunc = pd.qcut if quantiles else pd.cut
        plotDF['category'] = cutFunc(resultSeries, len(outputLabels), labels=outputLabels)

        colormap = 'rainbow'
        alpha = 0.4

        # color = [
        #     [0.8, 0.0, 0.1, alpha],
        #     [0.0, 0.8, 0.1, alpha],
        #     [0.1, 0.1, 0.8, alpha],
        # ]
        parallel_coordinates(plotDF, 'category', cols=cols, alpha=alpha,
                             #color=color,
                             colormap=colormap,
                             )
        fig = plt.gcf()
        fig.canvas.draw()       # so that ticks / labels are generated

        if rotation is not None:
            plt.xticks(rotation=rotation)

        # Labels can come out as follows for, say, 4 bins:
        # [u'', u'0.0', u'0.5', u'1.0', u'1.5', u'2.0', u'2.5', u'3.0', u'']
        # We eliminate the "x.5" labels by substituting '' and convert the remaining
        # numerical values to integers (i.e., eliminating ".0")
        def _fixTick(text):
            if inputBins:
                return '' if (not text or text.endswith('.5')) else str(int(float(text)))

            # If not binning, just show values on Y-axis
            return text

        locs, ylabels = plt.yticks()
        ylabels = [_fixTick(t._text) for t in ylabels]
        plt.yticks(locs, ylabels)

        if extra:
            printExtraText(fig, extra, loc='top', color='lightgrey', weight='ultralight', fontsize='xx-small')

        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        if title:
            plt.title(title)

        if show:
            plt.show()

        if filename:
            _logger.debug("Saving parallel coordinates plot to %s" % filename)
            plt.savefig(filename, bbox_inches='tight')

        plt.close(fig)


def analyzeSimulationNew(args):
    '''
    Analyze a simulation by reading parameters and results from the database.
    '''
    simId       = args.simId
    expNames    = args.expName
    plotHist    = args.plot
    stats       = args.stats
    importance  = args.importance
    groups      = args.groups
    plotInputs  = args.plotInputs
    convergence = args.convergence
    resultName  = args.resultName
    xlabel      = args.xlabel
    inputsFile  = args.exportInputs
    resultFile  = args.resultFile
    exportEMA   = args.exportEMA
    minimum     = args.min
    maximum     = args.max
    parallel    = args.parallel

    anaObj = Analysis(args.simId, args.scenarioNames, args.resultNames, args.limit)
    trials = anaObj.trials

    if inputsFile:
        anaObj.exportInputs(inputsFile)

    if not (expNames and resultName):
        raise PygcamMcsUserError("expName and resultName must be specified")

    expList = expNames.split(',')
    resultList = resultName.split(',')

    if resultFile:
        anaObj.exportResults(resultFile, scenarioList=expList, resultList=resultList)
        return

    inputDF = anaObj.getInputs()
    inputRows, inputCols = inputDF.shape

    if exportEMA:
        saveForEMA(simId, expList, resultList, inputDF, exportEMA)
        return

    if plotInputs:
        plotInputDistributions()

    if not (importance or groups or plotHist or convergence or stats):
        return

    resultsDict = anaObj.getResults(scenarioList=expList, resultList=[resultName])

    for expName in expList:
        resultDF = resultsDict[expName]

        if resultDF is None:
            raise PygcamMcsSystemError('analyze: No results for simId=%d, expName=%s' % (simId, expName))

        if maximum is not None:
            before = resultDF.shape[0]
            resultDF = resultDF[resultDF[resultName] <= maximum]
            after = resultDF.shape[0]
            _logger.debug('Applying maximum value (%f) eliminated %d rows.', maximum, before - after)

        if minimum is not None:
            before = resultDF.shape[0]
            resultDF = resultDF[resultDF[resultName] >= minimum]
            after = resultDF.shape[0]
            _logger.debug('Applying minimum value (%f) eliminated %d rows.', minimum, before - after)

        resultSeries = resultDF[resultName]
        numResults = resultSeries.count()

        if plotHist:
            plotOutputDistribution(simId, expName, resultSeries, resultName, xlabel, trials)

        if stats:
            printStats(resultSeries)    # TBD: use resultSeries.describe() instead?

        if convergence:
            plotConvergence(simId, expName, resultName, resultSeries, show=False, save=True)

        if (importance or groups or inputsFile) and (numResults != trials or numResults != inputRows):
            _logger.info("SimID %d has %d trials, %d input rows, and %d results",
                         simId, trials, inputRows, numResults)

        if importance or groups or parallel:
            inputsWithResults = inputDF.ix[resultDF.index]

            # Drop any inputs with names ending in '-linked' since these are an artifact
            linked = filter(lambda s: s.endswith('-linked'), inputsWithResults.columns)
            if linked:
                inputsWithResults.drop(linked, axis=1, inplace=True)

            extraText = 'SimId=%d, Exp=%s, Trials=%d/%d' % (simId, expName, numResults, trials)

            # TBD: eliminate 'groups' and call it importance instead.
            if importance or groups:
                basename = '%s-s%02d-%s-sensitivity' % (expName, simId, resultName)
                filename = makePlotPath(basename, simId)

                anaObj.plotUncertaintyImportance(inputsWithResults, resultSeries,
                                                 filename=filename, extra=extraText)

            # TBD: drop this from 'analyze' command?
            if parallel:
                basename = '%s-s%02d-%s-parallel' % (expName, simId, resultName)
                filename = makePlotPath(basename, simId)

                # TBD: plot parallel coordinates figure
                anaObj.plotParallelCoordinates(inputsWithResults, resultSeries,
                                               filename=filename, extra=extraText)


def analyzeSimulation(args):
    '''
    Analyze a simulation by reading parameters and results from the database.
    '''
    simId       = args.simId
    expNames    = args.expName
    plotHist    = args.plot
    stats       = args.stats
    importance  = args.importance
    groups      = args.groups
    plotInputs  = args.plotInputs
    convergence = args.convergence
    resultName  = args.resultName
    limit       = args.limit
    xlabel      = args.xlabel
    inputsFile  = args.exportInputs
    resultFile  = args.resultFile
    exportEMA   = args.exportEMA
    minimum     = args.min
    maximum     = args.max

    inputDF = None

    db = getDatabase()
    trials = db.getTrialCount(simId) if limit <= 0 else limit
    if not trials:
        raise PygcamMcsUserError('No trials were found for simId %d' % simId)

    # TBD: instead, have commands requiring inputs call
    # inputs are shared across experiments, so gather these before looping over experiments
    if plotInputs or inputsFile or importance or groups or exportEMA:
        inputDF = readParameterValues(simId, trials)
        inputRows, inputCols = inputDF.shape
        _logger.info("Each trial has %d parameters", inputCols)

    if inputsFile:
        exportInputs(inputsFile, inputDF)

    expList = expNames and expNames.split(',')
    if (not (expList and expList[0] and resultName)):
        raise PygcamMcsUserError("expName and resultName must be specified")

    if resultFile:
        resultList = resultName.split(',')
        exportResults(simId, resultList, expList, resultFile)
        return

    if exportEMA:
        resultList = resultName.split(',')
        saveForEMA(simId, expList, resultList, inputDF, exportEMA)
        return

    if plotInputs:
        plotInputDistributions(simId, inputDF)

    if not (importance or groups or plotHist or convergence or stats):
        return

    for expName in expList:
        resultDF = db.getOutValues(simId, expName, resultName, limit=limit)
        if resultDF is None:
            raise PygcamMcsSystemError('analyzeSimulation: No results for simId=%d, expName=%s, resultName=%s' % (simId, expName, resultName))

        if maximum is not None:
            before = resultDF.shape[0]
            resultDF = resultDF[resultDF[resultName] <= maximum]
            after  = resultDF.shape[0]
            _logger.debug('Applying maximum value (%f) eliminated %d rows.', maximum, before - after)

        if minimum is not None:
            before = resultDF.shape[0]
            resultDF = resultDF[resultDF[resultName] >= minimum]
            after = resultDF.shape[0]
            _logger.debug('Applying minimum value (%f) eliminated %d rows.', minimum, before - after)

        resultSeries = resultDF[resultName]
        numResults = resultSeries.count()

        if plotHist:
            plotOutputDistribution(simId, expName, resultSeries, resultName, xlabel, trials)

        if stats:
            printStats(resultSeries)

        if convergence:
            plotConvergence(simId, expName, resultName, resultSeries, show=False, save=True)

        if (importance or groups or inputsFile) and (numResults != trials or numResults != inputRows):
            _logger.info("SimID %d has %d trials, %d input rows, and %d results", simId, trials, inputRows, numResults)

        if importance or groups:
            inputsWithResults = inputDF.ix[resultDF.index]

            # Drop any inputs with names ending in '-linked' since these are an artifact
            # Column names can look like 'foobar[0][34]', so we strip off indexing part.
            def _isLinked(colname):
                pos = colname.find('[')
                colname = colname if pos < 0 else colname[0:pos]
                return colname.endswith('-linked')

            linked = filter(_isLinked, inputsWithResults.columns)
            if linked:
                inputsWithResults.drop(linked, axis=1, inplace=True)

            spearman = spearmanCorrelation(inputsWithResults, resultSeries)

            data = pd.DataFrame(spearman)
            data['normalized'] = normalizeSeries(spearman ** 2)
            data['sign'] = 1
            negatives = (data.spearman < 0)
            data.ix[negatives, 'sign'] = -1
            data['value'] = data.normalized * data.sign     # normalized squares with signs restored

            if importance:
                plotSensitivityResults(resultName, data, limit=15,
                                        filename=makePlotPath('%s-s%02d-%s-ind' % (expName, simId, resultName), simId),
                                        extra='SimId=%d, Exp=%s, Trials=%d/%d' % (simId, expName, numResults, trials))

            if groups:
                plotGroupSensitivityResults(resultName, data, limit=15,
                                             filename=makePlotPath('%s-s%02d-%s-grp' % (expName, simId, resultName), simId),
                                             extra='SimId=%d, Exp=%s, Trials=%d/%d' % (simId, expName, numResults, trials))



#!/usr/bin/env python
"""
.. Workspace sub-command

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016  Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from __future__ import print_function
from ..subcommand import SubcommandABC


def driver(args, tool):
    # lazy imports to avoid loading anything that's not used by gcamtool
    import os
    import subprocess

    from ..config import getParam
    from ..error import CommandlineError
    from ..scenarioSetup import createSandbox
    from ..log import getLogger
    from ..utils import removeTreeSafely, pathjoin

    _logger = getLogger(__name__)

    project = args.projectName or getParam('GCAM.DefaultProject')

    if not project:
        raise CommandlineError("sandbox: must specify project name or set config parameter GCAM.DefaultProject")

    if not (args.scenario or args.groupDir):
        raise CommandlineError("sandbox: must specify scenario and/or group name")

    sandboxProjectDir = getParam('GCAM.SandboxProjectDir')
    sandbox = pathjoin(sandboxProjectDir, args.groupDir, args.scenario)

    sandbox = os.path.normpath(os.path.abspath(os.path.expanduser(sandbox)))     # handle ~ in pathname

    if args.path:
        print(sandbox)

    execute = not args.noExecute

    if args.recreate:
        args.delete = args.create = True

    if args.delete:
        _logger.info('Removing ' + sandbox)
        try:
            if execute:
                if os.path.islink(sandbox):
                    os.remove(sandbox)
                else:
                    removeTreeSafely(sandbox)
            else:
                print("Would remove", sandbox)
        except Exception as e:
            _logger.warn("Can't remove '%s': %s" % (sandbox, e))

    if args.create:
        if execute:
            _logger.info('Creating ' + sandbox)
            createSandbox(sandbox)
        else:
            print("Would create", sandbox)

    if args.run:
        cmdStr = 'cd ' + sandbox + '; ' + args.run
        if execute:
            _logger.info(cmdStr)
            os.chdir(sandbox)
            subprocess.call(args.run, shell=True)
        else:
            print("Would run:", cmdStr)

class SandboxCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Perform operations on a sandbox.'''}
        super(SandboxCommand, self).__init__('sandbox', subparsers, kwargs, group='utils')

    def addArgs(self, parser):
        parser.add_argument('--create', action='store_true',
                            help='''Create the identified sandbox. If used with --delete,
                            the deletion occurs first.''')

        parser.add_argument('--delete', action='store_true',
                            help='''Delete the identified sandbox' If used with --create, the
                            deletion occurs first.''')

        parser.add_argument('--recreate', action='store_true',
                            help='''Recreate the identified sandbox. Equivalent to using the
                            --delete and --create options together.''')

        parser.add_argument('-g', '--groupDir', default='', metavar='NAME',
                            help='''The name of the scenario group subdir''')

        parser.add_argument('-n', '--noExecute', action='store_true',
                            help='''Print the command that would be executed by --run, but
                            don't execute it.''')

        parser.add_argument('-p', '--path', action='store_true',
                            help='''Print the absolute path to the identified sandbox.''')

        parser.add_argument('-r', '--run', metavar='CMD',
                            help='''Run the given command in the identified sandbox.''')

        parser.add_argument('-s', '--scenario', default='',
                            help='''The scenario for the computed sandbox root.''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



from __future__ import print_function
import subprocess as subp
import time

from pygcam.log import getLogger
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, Event

_logger = getLogger(__name__)

class Terminal(object):

    counter = 0     # to create unique ids

    def __init__(self, updateSeconds=2, toConsole=False):
        self.status = None
        self.proc   = None
        self.page   = None
        self.queue  = None
        self.toConsole = toConsole

        self.counter += 1
        self.intervalId = intervalId = 'interval-%d' % self.counter
        self.terminalId = terminalId = 'terminal' # 'terminal-%d' % self.counter
        self.buttonId   = buttonId   = 'button-%d'   % self.counter

        self.running = False
        self.ms = 1000 * updateSeconds
        self.text = ''
        self.layout = html.Div([
            # Non-displayed divs are useful for chaining events & inputs
            html.Div(id='timer-div',  style={'display': 'none'}),
            html.Div(id='button-div', style={'display': 'none'}),

            dcc.Interval(id=intervalId, interval=1e6), # initially a long interval

            html.Button('Run', id=buttonId, className='shadow'),
            html.Br(),
            html.Div(children="[No process is running]",
                     id=terminalId,
                     className='terminal')
        ], className='centered')


    def registerCallbacks(self, app):
        terminalId = self.terminalId
        intervalId = self.intervalId
        buttonId   = self.buttonId

        # Add support for staying scrolled to bottom of 'terminal' div
        # app.scripts.append_script({'external_url': 'https://codepen.io/plevin/pen/MvpeNV.js'})

        # Since events aren't handled as well as inputs, we convert the
        # timer event into an input by setting the value in a hidden <div>.
        @app.callback(Output('timer-div', 'children'),
                      events=[Event(intervalId, 'interval')])
        def timerToInput():
            return str(time.time())     # because it changes each call

        @app.callback(Output(terminalId, 'children'),
                      [Input('timer-div', 'children')])
        def updateTerminal(timer):
            # print("updateTerminal(%s)" % timer)
            if not self.proc:
                return self.text

            newText = ''

            # Loop while there is data to read
            q = self.queue
            while q.qsize() > 0:
                buf = q.get_nowait()
                if buf == '':
                    break

                newText += buf

            self.text += newText

            self.status = self.proc.poll()
            self.running = self.status is None
            if not self.running:
                self.text += "\n[Process exited]\n"
                self.proc = self.fd = None

            return self.text


        @app.callback(Output('button-div', 'children'),     # writing to this is a no-op
                      [Input(buttonId, 'n_clicks')])
        def processClick(clicks):
            if clicks is None:
                return ''

            if self.running:
                self.stopCommand()
            else:
                command = self.page.getCommand()
                self.runCommand(command)

            return str(time.time())

        # Set button text to match running state, on click or timer.
        @app.callback(Output(buttonId, 'children'),
                      [Input('button-div', 'children'),
                       Input('timer-div', 'children')])
        def clicks(buttonInfo, timerInfo):
            # print("clicks: button:%s, timer:%s" % (buttonInfo, timerInfo))
            return 'Stop' if self.running else 'Run'


        # Set a short interval timer when a proc is running, otherwise
        # set it to an hour (effectively disabling it.)
        @app.callback(Output(intervalId, 'interval'),
                      [Input('button-div', 'children'),
                       Input('timer-div', 'children')])
        def setTimer(buttonInfo, timerInfo):
            # print("setTimer: button:%s, timer:%s" % (buttonInfo, timerInfo))
            oneHour = 60 * 60 * 1000
            ms = self.ms if self.running else oneHour
            return ms

    def setPage(self, page):
        # print("Setting page to %s" % page)
        self.page = page

    def stopCommand(self):
        import time
        proc = self.proc

        if proc:
            pid = proc.pid
            self.text += "\n[Sending SIGTERM to process %d]\n" % pid
            proc.terminate()
            time.sleep(2)
            if proc.poll() is None:  # follow with SIGKILL if process is persistent
                time.sleep(3)
                self.text += "[Sending SIGKILL to process %d]\n" % pid
                proc.kill()

            proc.wait()
            self.proc = self.fd = None
            self.text += "[Terminated process %d]\n" % pid

        self.running = False

    def runCommand(self, command):
        from ..utils import queueForStream

        if self.running:
            _logger.warning('Already running a command')
            return

        self.status = None
        self.text = '$ ' + command + '\n'

        _logger.info('Running "%s"' % command)
        if self.toConsole:
            subp.call(command, shell=True)
        else:
            self.proc = subp.Popen(command, stdout=subp.PIPE, stderr=subp.STDOUT, shell=True,   # for now...
                                   bufsize=1,    # line buffered
                                   universal_newlines=True, cwd=None, env=None)

            self.queue = queueForStream(self.proc.stdout)
            self.running = True




'''
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
#
# Constants shared across modules
#

# TBD: move to gcamtool package
LOCAL_XML_NAME = 'local-xml'
DYN_XML_NAME   = 'dyn-xml'
XML_SRC_NAME   = 'xmlsrc'

# These are the "standard" unmanaged classes. 'OtherArableLand' can also be protected.
UnmanagedLandClasses = ['UnmanagedPasture', 'UnmanagedForest', 'Shrubland', 'Grassland']

# TBD: keep in pygcam package
DEFAULT_TIMESTEP = 5    # 5 years

NUM_AEZS = 18

GCAM_32_REGIONS = [
    'Africa_Eastern',
    'Africa_Northern',
    'Africa_Southern',
    'Africa_Western',
    'Argentina',
    'Australia_NZ',
    'Brazil',
    'Canada',
    'Central America and Caribbean',
    'Central Asia',
    'China',
    'Colombia',
    'EU-12',
    'EU-15',
    'Europe_Eastern',
    'Europe_Non_EU',
    'European Free Trade Association',
    'India',
    'Indonesia',
    'Japan',
    'Mexico',
    'Middle East',
    'Pakistan',
    'Russia',
    'South Africa',
    'South America_Northern',
    'South America_Southern',
    'South Asia',
    'South Korea',
    'Southeast Asia',
    'Taiwan',
    'USA'
]



'''
.. Created on: 2/26/15

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from ..subcommand import SubcommandABC

class GcamCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Run GCAM for the indicated configFile, scenario, or workspace.'''}
        super(GcamCommand, self).__init__('gcam', subparsers, kwargs, group='project', label='GCAM')

    def addArgs(self, parser):
        # parser.add_argument('-c', '--copyWorkspace', action='store_true',
        #                     help='''Create a copy of the source workspace in the location specified
        #                     by GCAM.''')

        parser.add_argument('-C', '--configFile',
                            help='''Specify the one or more GCAM configuration filenames, separated by commas.
                            If multiple configuration files are given, the are run in succession in the
                            same "job" on the cluster.''')

        parser.add_argument('-f', '--forceCreate', action='store_true',
                            help='''Re-create the workspace, even if it already exists.''')

        parser.add_argument('-g', '--groupDir', default='',
                            help='The scenario group directory name, if any.')

        parser.add_argument('-n', '--noRun', action='store_true',
                            help='''Don't run GCAM; just print the command that would be run.''')

        parser.add_argument('-r', '--refWorkspace',
                            help='''The reference workspace to use to create the new sandbox. This is
                            used only if the indicated or implied workspace doesn't exist. Defaults
                            to the value of GCAM.RefWorkspace.''')

        parser.add_argument('-s', '--scenario', default='',
                            help='''The scenario to run.''')

        parser.add_argument('-S', '--scenariosDir', default='',
                            help='''Specify the directory holding scenario files. Default is the value of
                            config variable GCAM.ScenariosDir, if set, otherwise it's the current directory.''')

        parser.add_argument('-w', '--workspace',
                            help='''Specify the path to the GCAM workspace to use. If it doesn't exist, the
                            named workspace will be created. If not specified on the command-line, the path
                            constructed as {GCAM.SandboxDir}/{scenario} is used.''')

        parser.add_argument('-W', '--noWrapper', action='store_true',
                            help='''Do not run gcam within a wrapper that detects errors as early as possible
                            and terminates the model run. By default, the wrapper is used.''')
        return parser

    def run(self, args, tool):
        from ..gcam import gcamMain
        gcamMain(args)
